# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Function transformations like Fourier or wavelet transforms."""

from . import backends, util
from .backends import PYFFTW_AVAILABLE, PYWT_AVAILABLE
from .fourier import *
from .wavelet import *
from .deform import *

__all__ = ()
__all__ += fourier.__all__
__all__ += wavelet.__all__
__all__ += deform.__all__
__all__ += ("PYFFTW_AVAILABLE", "PYWT_AVAILABLE")



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Tests for the Filtered Back-Projection."""

import numpy as np
import pytest

import odl
from odl.applications.tomo.util.testutils import (
    skip_if_no_astra, skip_if_no_astra_cuda, skip_if_no_skimage, skip_if_no_pytorch)
from odl.core.util.testutils import simple_fixture

# --- pytest fixtures --- #

### The threshold values of L2-norms and element-wise maximum differences
### are measured with the "m" and "n_angles" parameters below.
### If those are changed, the tests probably won't pass.
m = 500
n_angles = 1000

pytestmark = pytest.mark.suite('largescale')

impl = simple_fixture(
    name='impl',
    params=[pytest.param('astra_cpu', marks=skip_if_no_astra),
            pytest.param('astra_cuda', marks=skip_if_no_astra_cuda),
            pytest.param('skimage', marks=skip_if_no_skimage)]
)

geometry_params = ['par2d', 'par3d', 'cone2d', 'cone3d', 'helical']
geometry_ids = [" geometry='{}' ".format(p) for p in geometry_params]


@pytest.fixture(scope='module', ids=geometry_ids, params=geometry_params)
def geometry(request):
    geom = request.param

    if geom == 'par3d':
        apart = odl.uniform_partition(0, np.pi, n_angles)
        dpart = odl.uniform_partition([-30, -30], [30, 30], (m, m))
        return odl.applications.tomo.Parallel3dAxisGeometry(apart, dpart)
    elif geom == 'cone3d':
        apart = odl.uniform_partition(0, 2 * np.pi, n_angles)
        dpart = odl.uniform_partition([-60, -60], [60, 60], (m, m))
        return odl.applications.tomo.ConeBeamGeometry(apart, dpart,
                                         src_radius=200, det_radius=100)
    elif geom == 'helical':
        apart = odl.uniform_partition(0, 8 * 2 * np.pi, n_angles)
        dpart = odl.uniform_partition([-30, -3], [30, 3], (m, m))
        return odl.applications.tomo.ConeBeamGeometry(apart, dpart, pitch=5.0,
                                         src_radius=200, det_radius=100)
    else:
        raise ValueError('geom not valid')
    
geometry_type = simple_fixture(
    'geometry_type',
    ['par2d', 'par3d', 'cone2d', 'cone3d']
)

### The last two numbers are "hand-picked" by looking the FBP reconstructions
### for different scenarios. The first number is the L2-norm of the difference
### between Shepp-Logan phantom and the FBP reconstruction.
### The second number is maximum distance between element-wise comparison of 
### the Shepp-Logan phantom and the FBP reconstruction.
projectors = []
projectors.extend(
    (pytest.param(proj_cfg, marks=skip_if_no_astra_cuda)
     for proj_cfg in [
                      'par3d astra_cuda numpy cpu Ram-Lak 3.46 0.20',
                      'par3d astra_cuda numpy cpu Shepp-Logan 4.01 0.19',
                      'par3d astra_cuda numpy cpu Cosine 5.68 0.20',
                      'par3d astra_cuda numpy cpu Hamming 6.89 0.23',
                      'par3d astra_cuda numpy cpu Hann 7.25 0.23',
                      'cone3d astra_cuda numpy cpu Ram-Lak 5.41 0.23',
                      'cone3d astra_cuda numpy cpu Shepp-Logan 6.24 0.23',
                      'cone3d astra_cuda numpy cpu Cosine 8.10 0.26',
                      'cone3d astra_cuda numpy cpu Hamming 9.60 0.30',
                      'cone3d astra_cuda numpy cpu Hann 10.00 0.32',
                      'helical astra_cuda numpy cpu Ram-Lak 42.68 0.72',
                      'helical astra_cuda numpy cpu Shepp-Logan 41.50 0.69',
                      'helical astra_cuda numpy cpu Cosine 39.54 0.62',
                      'helical astra_cuda numpy cpu Hamming 38.45 0.58',
                      'helical astra_cuda numpy cpu Hann 38.16 0.56'
                      ])
)

projectors.extend(
    (pytest.param(proj_cfg, marks=[skip_if_no_astra, skip_if_no_pytorch])
     for proj_cfg in [
                      'par3d astra_cuda pytorch cuda:0 Ram-Lak 3.46 0.20',
                      'par3d astra_cuda pytorch cuda:0 Shepp-Logan 4.01 0.19',
                      'par3d astra_cuda pytorch cuda:0 Cosine 5.68 0.20',
                      'par3d astra_cuda pytorch cuda:0 Hamming 6.87 0.23',
                      'par3d astra_cuda pytorch cuda:0 Hann 7.25 0.23',
                      'cone3d astra_cuda pytorch cuda:0 Ram-Lak 5.41 0.23',
                      'cone3d astra_cuda pytorch cuda:0 Shepp-Logan 6.24 0.23',
                      'cone3d astra_cuda pytorch cuda:0 Cosine 8.01 0.26',
                      'cone3d astra_cuda pytorch cuda:0 Hamming 9.60 0.30',
                      'cone3d astra_cuda pytorch cuda:0 Hann 10.00 0.32',
                      'helical astra_cuda pytorch cuda:0 Ram-Lak 42.68 0.72',
                      'helical astra_cuda pytorch cuda:0 Shepp-Logan 41.50 0.69',
                      'helical astra_cuda pytorch cuda:0 Cosine 39.54 0.62',
                      'helical astra_cuda pytorch cuda:0 Hamming 38.45 0.58',
                      'helical astra_cuda pytorch cuda:0 Hann 38.16 0.56'])
)



projector_ids = [
    " geom='{}' - astra_impl='{}' - tspace_impl='{}' - tspace_device='{}' - filter='{}' - L2_value='{}' - max_dist='{}'".format(*p.values[0].split())
    for p in projectors
]


@pytest.fixture(scope='module', params=projectors, ids=projector_ids)
def projector(request):
    ### Change of the "n" variable can affect the passing requirements.
    n = 100
    dtype = 'float32'
    geom, astra_impl, tspace_impl, tspace_device, filter, L2_value, max_dist = request.param.split()
    
    apart = odl.uniform_partition(0, 2 * np.pi, n_angles)

    if geom == 'par2d':
        # Reconstruction space
        reco_space = odl.uniform_discr([-20] * 2, [20] * 2, [n] * 2,
                                       dtype=dtype, impl=tspace_impl, device=tspace_device)
        # Geometry
        dpart = odl.uniform_partition(-30, 30, m)
        geom = odl.applications.tomo.Parallel2dGeometry(apart, dpart)

    elif geom == 'par3d':
        # Reconstruction space
        reco_space = odl.uniform_discr([-20] * 3, [20] * 3, [n] * 3,
                                       dtype=dtype, impl=tspace_impl, device=tspace_device)

        # Geometry
        dpart = odl.uniform_partition([-30] * 2, [30] * 2, [m] * 2)
        geom = odl.applications.tomo.Parallel3dAxisGeometry(apart, dpart)

    elif geom == 'cone2d':
        # Reconstruction space
        reco_space = odl.uniform_discr([-20] * 2, [20] * 2, [n] * 2,
                                       dtype=dtype, impl=tspace_impl, device=tspace_device)
        # Geometry
        dpart = odl.uniform_partition(-30, 30, m)
        geom = odl.applications.tomo.FanBeamGeometry(apart, dpart, src_radius=200,
                                        det_radius=100)

    elif geom == 'cone3d':
        # Reconstruction space
        reco_space = odl.uniform_discr([-20] * 3, [20] * 3, [n] * 3,
                                       dtype=dtype, impl=tspace_impl, device=tspace_device)
        # Geometry
        dpart = odl.uniform_partition([-60] * 2, [60] * 2, [m] * 2)
        geom = odl.applications.tomo.ConeBeamGeometry(apart, dpart,
                                         src_radius=200, det_radius=100)

    elif geom == 'helical':
        # Reconstruction space
        reco_space = odl.uniform_discr([-20, -20, 0], [20, 20, 40],
                                       [n] * 3, dtype=dtype, impl=tspace_impl, device=tspace_device)
        # Geometry, overwriting angle partition
        apart = odl.uniform_partition(0, 8 * 2 * np.pi, n_angles)
        dpart = odl.uniform_partition([-30, -3], [30, 3], [m] * 2)
        geom = odl.applications.tomo.ConeBeamGeometry(apart, dpart, pitch=5.0,
                                         src_radius=200, det_radius=100)
    else:
        raise ValueError('geom not valid')
    
    # Ray transform
    return {'ray_trafo': odl.applications.tomo.RayTransform(reco_space, geom, impl=astra_impl, use_cache=False),
            'filter_type': filter,
            'L2_norm': L2_value,
            'max_distance': max_dist}
    # return (odl.applications.tomo.RayTransform(reco_space, geom, impl=astra_impl, use_cache=False), filter, L2_value, max_dist)


@pytest.fixture(scope='module',
                params=[True, False],
                ids=[' in-place ', ' out-of-place '])
def in_place(request):
    return request.param

def test_fpb(in_place, projector):
    """Test Filtered Back-Projection"""
    
    ray_trafo = projector['ray_trafo']
    filter = projector['filter_type']
    L2_value = float(projector['L2_norm'])
    max_dist = float(projector['max_distance'])

    rtol_norm = 0.05
    abs_dist = 0.05
    
    phantom = odl.core.phantom.shepp_logan(ray_trafo.domain, modified=True)
    # Filtered Back-Projection with different filters.
    fbp = odl.applications.tomo.fbp_op(ray_trafo, filter_type=filter, frequency_scaling=1.0)

    if in_place:
        proj = ray_trafo.range.zero()
        backproj = fbp.range.zero()
        fbp(ray_trafo(phantom, out=proj), out=backproj)
    else:
        proj = ray_trafo(phantom)
        backproj = fbp(proj)
    
    L2Norm = odl.functionals.default_functionals.L2Norm(fbp.range)
    assert L2Norm(phantom-backproj) == pytest.approx(L2_value, rel=L2_value*rtol_norm)
    assert odl.max(phantom-backproj) == pytest.approx(max_dist, abs=abs_dist)


if __name__ == '__main__':
    odl.core.util.test_file(__file__, ['-S', 'largescale'])



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Operators defined for tensor fields."""


import numpy as np
from math import prod

from odl.core.discr.discr_space import DiscretizedSpace
from odl.core.operator.tensor_ops import PointwiseTensorFieldOperator
from odl.core.space import ProductSpace
from odl.core.util import indent, signature_string, writable_array
from odl.core.array_API_support import asarray, get_array_and_backend

import warnings

__all__ = ('PartialDerivative', 'Gradient', 'Divergence', 'Laplacian')

_SUPPORTED_DIFF_METHODS = ('central', 'forward', 'backward')
_SUPPORTED_PAD_MODES = ('constant',
                        'symmetric', 'symmetric_adjoint',
                        'periodic',
                        'order0', 'order0_adjoint',
                        'order1', 'order1_adjoint',
                        'order2', 'order2_adjoint')

_ADJ_METHOD = {'central': 'central',
               'forward': 'backward',
               'backward': 'forward'}

_ADJ_PADDING = {'constant': 'constant',
                'symmetric': 'symmetric_adjoint',
                'symmetric_adjoint': 'symmetric',
                'periodic': 'periodic',
                'order0': 'order0_adjoint',
                'order0_adjoint': 'order0',
                'order1': 'order1_adjoint',
                'order1_adjoint': 'order1',
                'order2': 'order2_adjoint',
                'order2_adjoint': 'order2'}


class PartialDerivative(PointwiseTensorFieldOperator):

    """Calculate the discrete partial derivative along a given axis.

    Calls helper function `finite_diff` to calculate finite difference.
    Preserves the shape of the underlying grid.
    """

    def __init__(self, domain, axis, range=None, method='forward',
                 pad_mode='constant', pad_const=0):
        """Initialize a new instance.

        Parameters
        ----------
        domain : `DiscretizedSpace`
            Space of elements on which the operator can act.
        axis : int
            Axis along which the partial derivative is evaluated.
        range : `DiscretizedSpace`, optional
            Space of elements to which the operator maps, must have the
            same shape as ``domain``.
            For the default ``None``, the range is the same as ``domain``.
        method : {'forward', 'backward', 'central'}, optional
            Finite difference method which is used in the interior of the
            domain of ``f``.
        pad_mode : string, optional
            The padding mode to use outside the domain.

            ``'constant'``: Fill with ``pad_const``.

            ``'symmetric'``: Reflect at the boundaries, not doubling the
            outmost values.

            ``'periodic'``: Fill in values from the other side, keeping
            the order.

            ``'order0'``: Extend constantly with the outmost values
            (ensures continuity).

            ``'order1'``: Extend with constant slope (ensures continuity of
            the first derivative). This requires at least 2 values along
            each axis where padding is applied.

            ``'order2'``: Extend with second order accuracy (ensures continuity
            of the second derivative). This requires at least 3 values along
            the ``axis`` where padding is applied.

        pad_const : float, optional
            For ``pad_mode == 'constant'``, ``f`` assumes
            ``pad_const`` for indices outside the domain of ``f``

        Examples
        --------
        >>> f = np.array([[ 0.,  1.,  2.,  3.,  4.],
        ...               [ 0.,  2.,  4.,  6.,  8.]])
        >>> discr = odl.uniform_discr([0, 0], [2, 1], f.shape)
        >>> par_deriv = PartialDerivative(discr, axis=0, pad_mode='order1')
        >>> par_deriv(f)
        uniform_discr([ 0.,  0.], [ 2.,  1.], (2, 5)).element(
            [[ 0.,  1.,  2.,  3.,  4.],
             [ 0.,  1.,  2.,  3.,  4.]]
        )
        """
        if not isinstance(domain, DiscretizedSpace):
            raise TypeError(f"`domain` {domain} is not a DiscretizedSpace instance")

        if range is None:
            range = domain

        # Method is affine if nonzero padding is given.
        linear = not (pad_mode == 'constant' and pad_const != 0)
        super().__init__(domain, range, base_space=domain, linear=linear)
        self.axis = int(axis)
        self.dx = self.domain.cell_sides[axis]

        self.method, method_in = str(method).lower(), method
        if method not in _SUPPORTED_DIFF_METHODS:
            raise ValueError(f"`method` {method_in} not understood")

        self.pad_mode, pad_mode_in = str(pad_mode).lower(), pad_mode
        if pad_mode not in _SUPPORTED_PAD_MODES:
            raise ValueError(f"`pad_mode` {pad_mode_in} not understood")

        self.pad_const = self.domain.field.element(pad_const)

    def _call(self, x, out=None):
        """Calculate partial derivative of ``x``."""
        if out is None:
            out = self.range.element()

        # TODO: this pipes CUDA arrays through NumPy. Write native operator.
        with writable_array(out) as out_arr:
            finite_diff(x.asarray(), axis=self.axis, dx=self.dx,
                        method=self.method, pad_mode=self.pad_mode,
                        pad_const=self.pad_const, out=out_arr)
        return out

    def derivative(self, point=None):
        """Return the derivative operator.

        The partial derivative is usually linear, but in case the 'constant'
        ``pad_mode`` is used with nonzero ``pad_const``, the
        derivative is given by the derivative with 0 ``pad_const``.

        Parameters
        ----------
        point : `domain` `element-like`, optional
            The point to take the derivative in. Does not change the result
            since the operator is affine.
        """
        if self.pad_mode == 'constant' and self.pad_const != 0:
            return PartialDerivative(self.domain, self.axis, self.range,
                                     self.method, self.pad_mode, 0)
        else:
            return self

    @property
    def adjoint(self):
        """Return the adjoint operator."""
        if not self.is_linear:
            raise ValueError(
                f"operator with nonzero pad_const ({self.pad_const}) is not linear and has no adjoint"
            )

        return -PartialDerivative(self.range, self.axis, self.domain,
                                  _ADJ_METHOD[self.method],
                                  _ADJ_PADDING[self.pad_mode],
                                  self.pad_const)

    def __repr__(self):
        """Return ``repr(self)``."""
        posargs = [self.domain]
        optargs = [('axis', self.axis, None),
                   ('range', self.range, self.domain),
                   ('method', self.method, 'forward'),
                   ('pad_mode', self.pad_mode, 'constant'),
                   ('pad_const', self.pad_const, 0)]
        inner_str = signature_string(posargs, optargs,
                                     sep=',\n', mod=['!r', ''])
        return f"{self.__class__.__name__}(\n{indent(inner_str)}\n)"

    def __str__(self):
        """Return ``str(self)``."""
        dom_ran_str = "\n-->\n".join([repr(self.domain), repr(self.range)])
        return f"{self.__class__.__name__}:\n{indent(dom_ran_str)}"


class Gradient(PointwiseTensorFieldOperator):
    """Spatial gradient operator for `DiscretizedSpace` spaces.

    Calls helper function `finite_diff` to calculate each component of the
    resulting product space element. For the adjoint of the `Gradient`
    operator, zero padding is assumed to match the negative `Divergence`
    operator
    """

    def __init__(self, domain=None, range=None, method='forward',
                 pad_mode='constant', pad_const=0):
        """Initialize a new instance.

        Zero padding is assumed for the adjoint of the `Gradient`
        operator to match negative `Divergence` operator.

        Parameters
        ----------
        domain : `DiscretizedSpace`, optional
            Space of elements which the operator acts on.
            This is required if ``range`` is not given.
        range : power space of `DiscretizedSpace`, optional
            Space of elements to which the operator maps.
            This is required if ``domain`` is not given.
        method : {'forward', 'backward', 'central'}, optional
            Finite difference method to be used.
        pad_mode : string, optional
            The padding mode to use outside the domain.

            ``'constant'``: Fill with ``pad_const``.

            ``'symmetric'``: Reflect at the boundaries, not doubling the
            outmost values.

            ``'periodic'``: Fill in values from the other side, keeping
            the order.

            ``'order0'``: Extend constantly with the outmost values
            (ensures continuity).

            ``'order1'``: Extend with constant slope (ensures continuity of
            the first derivative). This requires at least 2 values along
            each axis where padding is applied.

            ``'order2'``: Extend with second order accuracy (ensures continuity
            of the second derivative). This requires at least 3 values along
            each axis.

        pad_const : float, optional
            For ``pad_mode == 'constant'``, ``f`` assumes
            ``pad_const`` for indices outside the domain of ``f``

        Examples
        --------
        Creating a Gradient operator:

        >>> dom = odl.uniform_discr([0, 0], [1, 1], (10, 20))
        >>> ran = odl.ProductSpace(dom, dom.ndim)  # 2-dimensional
        >>> grad_op = Gradient(dom)
        >>> grad_op.range == ran
        True
        >>> grad_op2 = Gradient(range=ran)
        >>> grad_op2.domain == dom
        True
        >>> grad_op3 = Gradient(domain=dom, range=ran)
        >>> grad_op3.domain == dom
        True
        >>> grad_op3.range == ran
        True

        Calling the operator:

        >>> data = np.array([[ 0., 1., 2., 3., 4.],
        ...                  [ 0., 2., 4., 6., 8.]])
        >>> discr = odl.uniform_discr([0, 0], [2, 5], data.shape)
        >>> f = discr.element(data)
        >>> grad = Gradient(discr)
        >>> grad_f = grad(f)
        >>> grad_f[0]
        uniform_discr([ 0.,  0.], [ 2.,  5.], (2, 5)).element(
            [[ 0.,  1.,  2.,  3.,  4.],
             [ 0., -2., -4., -6., -8.]]
        )
        >>> grad_f[1]
        uniform_discr([ 0.,  0.], [ 2.,  5.], (2, 5)).element(
            [[ 1.,  1.,  1.,  1., -4.],
             [ 2.,  2.,  2.,  2., -8.]]
        )

        Verify adjoint:

        >>> g = grad.range.element((data, data ** 2))
        >>> adj_g = grad.adjoint(g)
        >>> adj_g
        uniform_discr([ 0.,  0.], [ 2.,  5.], (2, 5)).element(
            [[ -0.,  -2.,  -5.,  -8., -11.],
             [ -0.,  -5., -14., -23., -32.]]
        )
        >>> g.inner(grad_f) / f.inner(adj_g)
        1.0
        """
        if domain is None and range is None:
            raise ValueError("either `domain` or `range` must be specified")

        if domain is None:
            try:
                domain = range[0]
            except TypeError:
                pass

        if range is None:
            range = ProductSpace(domain, domain.ndim)

        # Check range first since `domain` may end up to be `None` in
        # the case filtered out here (see above)
        if not isinstance(range, ProductSpace):
            raise TypeError(f"`range` {range} is not a `ProductSpace` instance")
        if not range.is_power_space:
            raise ValueError(f"`range` {range} is not a power space")

        if not isinstance(domain, DiscretizedSpace):
            raise TypeError(f"`domain` {domain} is not a `DiscretizedSpace` instance")

        if len(range) != domain.ndim:
            raise ValueError(
                f"`range` must be a power space of length n = {domain.ndim},with `n == domain.ndim`, got n = {len(range)} instead"
            )

        linear = not (pad_mode == 'constant' and pad_const != 0)
        super().__init__(domain, range, base_space=domain, linear=linear)

        self.method, method_in = str(method).lower(), method
        if method not in _SUPPORTED_DIFF_METHODS:
            raise ValueError(f"`method` {method_in} not understood")

        self.pad_mode, pad_mode_in = str(pad_mode).lower(), pad_mode
        if pad_mode not in _SUPPORTED_PAD_MODES:
            raise ValueError(f"`pad_mode` {pad_mode_in} not understood")

        self.pad_const = domain.field.element(pad_const)

    def _call(self, x, out=None):
        """Calculate the spatial gradient of ``x``."""
        x_arr = x.asarray()
        ndim = self.domain.ndim
        dx = self.domain.cell_sides

        if out is None:
            return self.range.element([
                     finite_diff(x_arr, axis=axis, dx=dx[axis], method=self.method,
                            pad_mode=self.pad_mode,
                            pad_const=self.pad_const,
                            )
                     for axis in range(ndim)])
        else:
            for axis in range(ndim):
                with writable_array(out[axis]) as out_arr:
                    finite_diff(x_arr, axis=axis, dx=dx[axis], method=self.method,
                                pad_mode=self.pad_mode,
                                pad_const=self.pad_const,
                                out=out_arr)
            return out

    def derivative(self, point=None):
        """Return the derivative operator.

        The gradient is usually linear, but in case the 'constant'
        ``pad_mode`` is used with nonzero ``pad_const``, the
        derivative is given by the Gradient with ``pad_const=0``.

        Parameters
        ----------
        point : `domain` element, optional
            The point to take the derivative in. Does not change the result
            since the operator is affine.
        """
        if self.pad_mode == 'constant' and self.pad_const != 0:
            return Gradient(self.domain, self.range, self.method,
                            pad_mode=self.pad_mode,
                            pad_const=0)
        else:
            return self

    @property
    def adjoint(self):
        """Adjoint of this operator.

        The adjoint is given by the negative `Divergence` with corrections for
        the method and padding.

        The Divergence is constructed from a ``space`` as a product space
        operator ``space^n --> space``, hence we need to provide the domain of
        this operator.
        """
        if not self.is_linear:
            raise ValueError(
                f"operator with nonzero pad_const ({self.pad_const}) is not linear and has no adjoint"
            )

        return - Divergence(domain=self.range, range=self.domain,
                            method=_ADJ_METHOD[self.method],
                            pad_mode=_ADJ_PADDING[self.pad_mode],
                            pad_const=self.pad_const)

    def __repr__(self):
        """Return ``repr(self)``."""
        posargs = [self.domain]
        optargs = [('range', self.range, self.domain ** self.domain.ndim),
                   ('method', self.method, 'forward'),
                   ('pad_mode', self.pad_mode, 'constant'),
                   ('pad_const', self.pad_const, 0)]
        inner_str = signature_string(posargs, optargs,
                                     sep=[',\n', ', ', ',\n'],
                                     mod=['!r', ''])
        return f"{self.__class__.__name__}(\n{indent(inner_str)}\n)"

    def __str__(self):
        """Return ``str(self)``."""
        dom_ran_str = "\n-->\n".join([repr(self.domain), repr(self.range)])
        return f"{self.__class__.__name__}:\n{indent(dom_ran_str)}"


class Divergence(PointwiseTensorFieldOperator):
    """Divergence operator for `DiscretizedSpace` spaces.

    Calls helper function `finite_diff` for each component of the input
    product space vector. For the adjoint of the `Divergence` operator to
    match the negative `Gradient` operator implicit zero is assumed.
    """

    def __init__(self, domain=None, range=None, method='forward',
                 pad_mode='constant', pad_const=0):
        """Initialize a new instance.

        Zero padding is assumed for the adjoint of the `Divergence`
        operator to match the negative `Gradient` operator.

        Parameters
        ----------
        domain : power space of `DiscretizedSpace`, optional
            Space of elements which the operator acts on.
            This is required if ``range`` is not given.
        range : `DiscretizedSpace`, optional
            Space of elements to which the operator maps.
            This is required if ``domain`` is not given.
        method : {'forward', 'backward', 'central'}, optional
            Finite difference method to be used
        pad_mode : string, optional
            The padding mode to use outside the domain.

            ``'constant'``: Fill with ``pad_const``.

            ``'symmetric'``: Reflect at the boundaries, not doubling the

            ``'periodic'``: Fill in values from the other side, keeping
            the order.

            ``'order0'``: Extend constantly with the outmost values
            (ensures continuity).

            ``'order1'``: Extend with constant slope (ensures continuity of
            the first derivative). This requires at least 2 values along
            each axis.

            ``'order2'``: Extend with second order accuracy (ensures continuity
            of the second derivative). This requires at least 3 values along
            each axis.

        pad_const : float, optional
            For ``pad_mode == 'constant'``, ``f`` assumes
            ``pad_const`` for indices outside the domain of ``f``

        Examples
        --------
        Initialize a Divergence opeator:

        >>> ran = odl.uniform_discr([0, 0], [3, 5], (3, 5))
        >>> dom = odl.ProductSpace(ran, ran.ndim)  # 2-dimensional
        >>> div = Divergence(dom)
        >>> div.range == ran
        True
        >>> div2 = Divergence(range=ran)
        >>> div2.domain == dom
        True
        >>> div3 = Divergence(domain=dom, range=ran)
        >>> div3.domain == dom
        True
        >>> div3.range == ran
        True

        Call the operator:

        >>> data = np.array([[0., 1., 2., 3., 4.],
        ...                  [1., 2., 3., 4., 5.],
        ...                  [2., 3., 4., 5., 6.]])
        >>> f = div.domain.element([data, data])
        >>> div_f = div(f)
        >>> print(div_f)
        [[  2.,   2.,   2.,   2.,  -3.],
         [  2.,   2.,   2.,   2.,  -4.],
         [ -1.,  -2.,  -3.,  -4., -12.]]

        Verify adjoint:

        >>> g = div.range.element(data ** 2)
        >>> adj_div_g = div.adjoint(g)
        >>> g.inner(div_f) / f.inner(adj_div_g)
        1.0
        """
        if domain is None and range is None:
            raise ValueError("either `domain` or `range` must be specified")

        if domain is None:
            domain = ProductSpace(range, range.ndim)

        if range is None:
            try:
                range = domain[0]
            except TypeError:
                pass

        # Check `domain` first since `range` may end up to be `None` in
        # the case filtered out here (see above)
        if not isinstance(domain, ProductSpace):
            raise TypeError(f"`domain` {domain} is not a `ProductSpace` instance")
        if not domain.is_power_space:
            raise ValueError(f"`domain` {domain} is not a power space")

        if not isinstance(range, DiscretizedSpace):
            raise TypeError(f"`range` {range} is not a `DiscretizedSpace` " "instance")

        if len(domain) != range.ndim:
            raise ValueError(
                f"`domain` must be a power space of length n = {range.ndim},with `n == range.ndim`, got n = {len(domain)} instead"
            )

        linear = not (pad_mode == 'constant' and pad_const != 0)
        super().__init__(domain, range, base_space=range, linear=linear)

        self.method, method_in = str(method).lower(), method
        if method not in _SUPPORTED_DIFF_METHODS:
            raise ValueError(f"`method` {method_in} not understood")

        self.pad_mode, pad_mode_in = str(pad_mode).lower(), pad_mode
        if pad_mode not in _SUPPORTED_PAD_MODES:
            raise ValueError(f"`pad_mode` {pad_mode_in} not understood")

        self.pad_const = range.field.element(pad_const)

    def _call(self, x, out=None):
        """Calculate the divergence of ``x``."""

        ndim = self.range.ndim
        dx = self.range.cell_sides

        backend = self.range.array_backend

        def directional_derivative(axis, out=None):
            return finite_diff( x[axis], axis=axis, dx=dx[axis]
                              , method=self.method, pad_mode=self.pad_mode
                              , pad_const=self.pad_const
                              , out=out )

        if out is None:
            result = directional_derivative(0)
            for axis in range(1,len(x)):
                result += directional_derivative(axis)

            return self.range.element(result)

        else:
            assert(backend.impl != 'pytorch')

            tmp = self.range.element().asarray()
            with writable_array(out) as out_arr:
                for axis in range(ndim):
                    directional_derivative(axis, out=tmp)
                    if axis == 0:
                        out_arr[:] = tmp
                    else:
                        out_arr += tmp
 
            return out

    def derivative(self, point=None):
        """Return the derivative operator.

        The Divergence is usually linear, but in case the 'constant'
        ``pad_mode`` is used with nonzero ``pad_const``, the
        derivative is given by the Divergence with ``pad_const=0``.

        Parameters
        ----------
        point : `domain` element, optional
            The point to take the derivative in. Does not change the result
            since the operator is affine.
        """
        if self.pad_mode == 'constant' and self.pad_const != 0:
            return Divergence(self.domain, self.range, self.method,
                              pad_mode=self.pad_mode, pad_const=0)
        else:
            return self

    @property
    def adjoint(self):
        """Adjoint of this operator.

        The adjoint is given by the negative `Gradient` with corrections for
        the method and padding.
        """
        if not self.is_linear:
            raise ValueError(
                f"operator with nonzero pad_const ({self.pad_const}) is not linear and has no adjoint"
            )

        return - Gradient(self.range, self.domain,
                          method=_ADJ_METHOD[self.method],
                          pad_mode=_ADJ_PADDING[self.pad_mode])

    def __repr__(self):
        """Return ``repr(self)``."""
        posargs = [self.domain]
        optargs = [('range', self.range, self.domain[0]),
                   ('method', self.method, 'forward'),
                   ('pad_mode', self.pad_mode, 'constant'),
                   ('pad_const', self.pad_const, 0)]
        inner_str = signature_string(posargs, optargs,
                                     sep=[',\n', ', ', ',\n'],
                                     mod=['!r', ''])
        return f"{self.__class__.__name__}(\n{indent(inner_str)}\n)"

    def __str__(self):
        """Return ``str(self)``."""
        dom_ran_str = "\n-->\n".join([repr(self.domain), repr(self.range)])
        return f"{self.__class__.__name__}:\n{indent(dom_ran_str)}"


class Laplacian(PointwiseTensorFieldOperator):
    """Spatial Laplacian operator for `DiscretizedSpace` spaces.

    Calls helper function `finite_diff` to calculate each component of the
    resulting product space vector.

    Outside the domain zero padding is assumed.
    """

    def __init__(self, domain, range=None, pad_mode='constant', pad_const=0):
        """Initialize a new instance.

        Parameters
        ----------
        domain : `DiscretizedSpace`
            Space of elements which the operator is acting on.
        pad_mode : string, optional
            The padding mode to use outside the domain.

            ``'constant'``: Fill with ``pad_const``.

            ``'symmetric'``: Reflect at the boundaries, not doubling the
            outmost values.

            ``'periodic'``: Fill in values from the other side, keeping
            the order.

            ``'order0'``: Extend constantly with the outmost values
            (ensures continuity).

            ``'order1'``: Extend with constant slope (ensures continuity of
            the first derivative). This requires at least 2 values along
            each axis.

            ``'order2'``: Extend with second order accuracy (ensures continuity
            of the second derivative). This requires at least 3 values along
            each axis where padding is applied.

        pad_const : float, optional
            For ``pad_mode == 'constant'``, ``f`` assumes
            ``pad_const`` for indices outside the domain of ``f``

        Examples
        --------
        >>> data = np.array([[ 0., 0., 0.],
        ...                  [ 0., 1., 0.],
        ...                  [ 0., 0., 0.]])
        >>> space = odl.uniform_discr([0, 0], [3, 3], [3, 3])
        >>> f = space.element(data)
        >>> lap = Laplacian(space)
        >>> lap(f)
        uniform_discr([ 0.,  0.], [ 3.,  3.], (3, 3)).element(
            [[ 0.,  1.,  0.],
             [ 1., -4.,  1.],
             [ 0.,  1.,  0.]]
        )
        """
        if not isinstance(domain, DiscretizedSpace):
            raise TypeError(f"`domain` {domain} is not a DiscretizedSpace instance")

        if range is None:
            range = domain

        super().__init__(domain, range, base_space=domain, linear=True)

        self.pad_mode, pad_mode_in = str(pad_mode).lower(), pad_mode
        if pad_mode not in _SUPPORTED_PAD_MODES:
            raise ValueError(f"`pad_mode` {pad_mode_in} not understood")
        if pad_mode in ('order1', 'order1_adjoint',
                        'order2', 'order2_adjoint'):
            # TODO: Add these pad modes
            raise ValueError(f"`pad_mode` {pad_mode_in} not implemented for Laplacian.")

        self.pad_const = self.domain.field.element(pad_const)

    def _call(self, x, out=None):
        """Calculate the spatial Laplacian of ``x``."""
        if out is None:
            out = self.range.zero()
        else:
            out.set_zero()

        out_arr = out.asarray()

        x_arr, backend = get_array_and_backend(x)
        tmp = backend.array_namespace.empty(
            shape=out.shape, dtype=out.dtype,device=x.device
            )

        ndim = self.domain.ndim
        dx = self.domain.cell_sides

        with writable_array(out) as out_arr:
            for axis in range(ndim):
                # TODO: this can be optimized
                finite_diff(x_arr, axis=axis, dx=dx[axis] ** 2,
                            method='forward',
                            pad_mode=self.pad_mode,
                            pad_const=self.pad_const, out=tmp)

                out_arr += tmp

                finite_diff(x_arr, axis=axis, dx=dx[axis] ** 2,
                            method='backward',
                            pad_mode=self.pad_mode,
                            pad_const=self.pad_const, out=tmp)

                out_arr -= tmp

        return out

    def derivative(self, point=None):
        """Return the derivative operator.

        The Laplacian is usually linear, but in case the 'constant'
        ``pad_mode`` is used with nonzero ``pad_const``, the
        derivative is given by the derivative with 0 ``pad_const``.

        Parameters
        ----------
        point : ``domain`` element, optional
            The point to take the derivative in. Does not change the result
            since the operator is affine.
        """
        if self.pad_mode == 'constant' and self.pad_const != 0:
            return Laplacian(self.domain, self.range,
                             pad_mode=self.pad_mode, pad_const=0)
        else:
            return self

    @property
    def adjoint(self):
        """Return the adjoint operator.

        The laplacian is self-adjoint, so this returns ``self``.
        """
        return Laplacian(self.range, self.domain,
                         pad_mode=self.pad_mode, pad_const=0)

    def __repr__(self):
        """Return ``repr(self)``."""
        posargs = [self.domain]
        optargs = [('range', self.range, self.domain ** self.domain.ndim),
                   ('pad_mode', self.pad_mode, 'constant'),
                   ('pad_const', self.pad_const, 0)]
        inner_str = signature_string(posargs, optargs,
                                     sep=[',\n', ', ', ',\n'],
                                     mod=['!r', ''])
        return f"{self.__class__.__name__}(\n{indent(inner_str)}\n)"

    def __str__(self):
        """Return ``str(self)``."""
        dom_ran_str = "\n-->\n".join([repr(self.domain), repr(self.range)])
        return f"{self.__class__.__name__}:\n{indent(dom_ran_str)}"


def _finite_diff_pure(f_arr, axis, dx=1.0, method='forward',
                pad_mode='constant', pad_const=0):
    """ Purely-functional version of `finite_diff`, using only primitives from
    the Python Array API. """

    ndim = f_arr.ndim

    if f_arr.shape[axis] < 2:
        raise ValueError(
            f"in axis {axis}: at least two elements required, got {f_arr.shape[axis]}"
        )

    if axis < 0:
        axis += ndim
    if not (0 <= axis < ndim):
        raise IndexError(f"`axis` {axis} outside the valid range 0 ... {ndim - 1}")

    dx, dx_in = float(dx), dx
    if dx <= 0 or not np.isfinite(dx):
        raise ValueError(f"`dx` must be positive, got {dx_in}")

    method, method_in = str(method).lower(), method
    if method not in _SUPPORTED_DIFF_METHODS:
        raise ValueError(f"`method` {method_in} was not understood")

    if pad_mode not in _SUPPORTED_PAD_MODES:
        raise ValueError(f"`pad_mode` {pad_mode} not understood")

    f_arr, backend = get_array_and_backend(f_arr)
    namespace = backend.array_namespace
    device = f_arr.device

    orig_shape = f_arr.shape

    if orig_shape[axis] < 2 and pad_mode == "order1":
        raise ValueError(
            f"size of array to small to use 'order1', needs at least 2 elements along axis {axis}."
        )
    if orig_shape[axis] < 3 and pad_mode == "order2":
        raise ValueError(
            f"size of array to small to use 'order2', needs at least 3 elements along axis {axis}."
        )

    # Reshape (in O(1)), so the axis of interest is the middle, all previous
    # axes are flattened into the batch dimension, and all subsequent axes flattened
    # into the final dimension. This allows a batched summing and indexing in only
    # one of the dimension, regardless of which axis is actually worked on.
    f_arr = f_arr.reshape([ prod(orig_shape[:axis])
                          , orig_shape[axis]
                          , prod(orig_shape[axis+1:])
                          ])

    # Interior of the domain of f
    if method == 'central':
        interior = (f_arr[:,2:] - f_arr[:,:-2]) / 2

    elif method == 'forward':
        interior = f_arr[:,2:] - f_arr[:,1:-1]

    elif method == 'backward':
        interior = f_arr[:,1:-1] - f_arr[:,:-2]

    corrections = {}

    # Boundaries
    if pad_mode == 'constant':
        # Assume constant value c for indices outside the domain of ``f``

        # With padding the method used on endpoints is the same as in the
        # interior of the domain of f

        if method == 'central':
            bnd_left = (f_arr[:,1] - pad_const) / 2.0
            bnd_right = (pad_const - f_arr[:,-2]) / 2.0

        elif method == 'forward':
            bnd_left = f_arr[:,1] - f_arr[:,0]
            bnd_right = pad_const - f_arr[:,-1]

        elif method == 'backward':
            bnd_left = f_arr[:,0] - pad_const
            bnd_right = f_arr[:,-1] - f_arr[:,-2]

    elif pad_mode == 'symmetric':
        # Values of f for indices outside the domain of f are replicates of
        # the edge values

        # With padding the method used on endpoints is the same as in the
        # interior of the domain of f

        if method == 'central':
            bnd_left = (f_arr[:,1] - f_arr[:,0]) / 2.0
            bnd_right = (f_arr[:,-1] - f_arr[:,-2]) / 2.0

        elif method == 'forward':
            bnd_left = f_arr[:,1] - f_arr[:,0]
            bnd_right = namespace.zeros_like(f_arr[:,-1])

        elif method == 'backward':
            bnd_left = namespace.zeros_like(f_arr[:,0])
            bnd_right = f_arr[:,-1] - f_arr[:,-2]

    elif pad_mode == 'symmetric_adjoint':
        # The adjoint case of symmetric

        if method == 'central':
            bnd_left = (f_arr[:,1] + f_arr[:,0]) / 2.0
            bnd_right = (-f_arr[:,-1] - f_arr[:,-2]) / 2.0

        elif method == 'forward':
            bnd_left = f_arr[:,1]
            bnd_right = -f_arr[:,-1]

        elif method == 'backward':
            bnd_left = f_arr[:,0]
            bnd_right = -f_arr[:,-2]

    elif pad_mode == 'periodic':
        # Values of f for indices outside the domain of f are replicates of
        # the edge values on the other side

        if method == 'central':
            bnd_left = (f_arr[:,1] - f_arr[:,-1]) / 2.0
            bnd_right = (f_arr[:,0] - f_arr[:,-2]) / 2.0

        elif method == 'forward':
            bnd_left = f_arr[:,1] - f_arr[:,0]
            bnd_right = f_arr[:,0] - f_arr[:,-1]

        elif method == 'backward':
            bnd_left = f_arr[:,0] - f_arr[:,-1]
            bnd_right = f_arr[:,-1] - f_arr[:,-2]

    elif pad_mode == 'order0':
        # Values of f for indices outside the domain of f are replicates of
        # the edge value.

        if method == 'central':
            bnd_left = (f_arr[:,1] - f_arr[:,0]) / 2.0
            bnd_right = (f_arr[:,-1] - f_arr[:,-2]) / 2.0

        elif method == 'forward':
            bnd_left = f_arr[:,1] - f_arr[:,0]
            bnd_right = namespace.zeros_like(f_arr[:,-1])

        elif method == 'backward':
            bnd_left = namespace.zeros_like(f_arr[:,0])
            bnd_right = f_arr[:,-1] - f_arr[:,-2]

    elif pad_mode == 'order0_adjoint':
        # Values of f for indices outside the domain of f are replicates of
        # the edge value.

        if method == 'central':
            bnd_left = (f_arr[:,0] + f_arr[:,1]) / 2.0
            bnd_right = -(f_arr[:,-1] + f_arr[:,-2]) / 2.0

        elif method == 'forward':
            bnd_left = f_arr[:,1]
            bnd_right = -f_arr[:,-1]

        elif method == 'backward':
            bnd_left = f_arr[:,0]
            bnd_right = -f_arr[:,-2]

    elif pad_mode == 'order1':
        # Values of f for indices outside the domain of f are linearly
        # extrapolated from the inside.

        # independent of ``method``

        bnd_left = f_arr[:,1] - f_arr[:,0]
        bnd_right = f_arr[:,-1] - f_arr[:,-2]

    elif pad_mode == 'order1_adjoint':
        # Values of f for indices outside the domain of f are linearly
        # extrapolated from the inside.

        if method == 'central':
            bnd_left = f_arr[:,0] + f_arr[:,1] / 2.0
            bnd_right = -f_arr[:,-1] - f_arr[:,-2] / 2.0

            corrections[1] = -f_arr[:,0] / 2.0
            corrections[-2] = f_arr[:,-1] / 2.0

        elif method == 'forward':
            bnd_left = f_arr[:,0] + f_arr[:,1]
            bnd_right = -f_arr[:,-1]

            corrections[1] = -f_arr[:,0]

        elif method == 'backward':
            bnd_left = f_arr[:,0]
            bnd_right = -f_arr[:,-1] - f_arr[:,-2]

            corrections[-2] = f_arr[:,-1]

    elif pad_mode == 'order2':
        # 2nd order edges

        bnd_left = -(3.0 * f_arr[:,0] - 4.0 * f_arr[:,1] + f_arr[:,2]) / 2.0
        bnd_right = (3.0 * f_arr[:,-1] - 4.0 * f_arr[:,-2] + f_arr[:,-3]) / 2.0

    elif pad_mode == 'order2_adjoint':
        # Values of f for indices outside the domain of f are quadratically
        # extrapolated from the inside.

        if method == 'central':
            bnd_left = 1.5 * f_arr[:,0] + 0.5 * f_arr[:,1]
            bnd_right = -1.5 * f_arr[:,-1] - 0.5 * f_arr[:,-2]

            corrections[1] = -1.5 * f_arr[:,0]
            corrections[2] = +0.5 * f_arr[:,0]
            corrections[-3] = -0.5 * f_arr[:,-1]
            corrections[-2] = +1.5 * f_arr[:,-1]

        elif method == 'forward':
            bnd_left = 1.5 * f_arr[:,0] + 1.0 * f_arr[:,1]
            bnd_right = -1.5 * f_arr[:,-1]

            corrections[1] = -2.0 * f_arr[:,0]
            corrections[2] = 0.5 * f_arr[:,0]
            corrections[-3] = -0.5 * f_arr[:,-1]
            corrections[-2] = 1.0 * f_arr[:,-1]

        elif method == 'backward':
            bnd_left = 1.5 * f_arr[:,0]
            bnd_right = -1.0 * f_arr[:,-2] - 1.5 * f_arr[:,-1]

            corrections[1] = -1.0 * f_arr[:,0]
            corrections[2] = 0.5 * f_arr[:,0]
            corrections[-3] = -0.5 * f_arr[:,-1]
            corrections[-2] = 2.0 * f_arr[:,-1]
    else:
        raise NotImplementedError("Unknown pad_mode")

    if corrections:
        uncorrected_start = max(corrections.keys()) + 1
        if uncorrected_start < 1:
            uncorrected_start = 1
        uncorrected_end = min(corrections.keys())
        if uncorrected_end >= 0:
            uncorrected_end = -1

        dummy_cols = namespace.zeros((f_arr.shape[0], 0, f_arr.shape[2]), device=device)

        correct_columns_l = namespace.stack(
                               [interior[:,i-1] + corrections[i]
                                for i in range(1, uncorrected_start)],
                               axis = 1) if uncorrected_start > 1 else dummy_cols
        correct_columns_r = namespace.stack(
                               [interior[:,i+1] + corrections[i]
                                for i in range(uncorrected_end, -1)],
                               axis = 1) if uncorrected_end < -1 else dummy_cols

        unaffected_width = f_arr.shape[1] + uncorrected_end - uncorrected_start
        if uncorrected_start <= 1:
            relevant_interior = interior[:, :uncorrected_end+1]
        elif uncorrected_end >= -1:
            relevant_interior = interior[:, uncorrected_start-1:]
        elif unaffected_width >= 0:
            relevant_interior = interior[:, uncorrected_start-1 : uncorrected_end+1]
        else:
            warnings.warn("Very short array, so the corrections from the left and right boundary"
                         +" overlap. This should generally not happen in practice (it indicates"
                         +" that the resolution is so low that the whole premise of approximating"
                         +" derivatives by finite differences fails)")
            overlap = -unaffected_width
            relevant_interior = (
                    correct_columns_l[:,-overlap:]
                    + correct_columns_r[:,:overlap]
                    - interior[:, uncorrected_end+1 : uncorrected_start-1])
            correct_columns_l = correct_columns_l[:,:-overlap]
            correct_columns_r = correct_columns_r[:,overlap:]

        result = namespace.concat( [ bnd_left[:,None],
                                     correct_columns_l,
                                     relevant_interior,
                                     correct_columns_r,
                                     bnd_right[:,None] ],
                                   axis = 1
                                 ) / dx
    else:
        result = namespace.concat([bnd_left[:,None], interior, bnd_right[:,None]], axis = 1) / dx

    return result.reshape(orig_shape)



def finite_diff(f, axis, dx=1.0, method='forward', out=None,
                pad_mode='constant', pad_const=0):
    """Calculate the partial derivative of ``f`` along a given ``axis``.

    In the interior of the domain of f, the partial derivative is computed
    using first-order accurate forward or backward difference or
    second-order accurate central differences.

    With padding the same method and thus accuracy is used on endpoints as
    in the interior i.e. forward and backward differences use first-order
    accuracy on edges while central differences use second-order accuracy at
    edges.

    Without padding one-sided forward or backward differences are used at
    the boundaries. The accuracy at the endpoints can then also be
    triggered by the edge order.

    The returned array has the same shape as the input array ``f``.

    Per default forward difference with dx=1 and no padding is used.

    Parameters
    ----------
    f : `array-like`
         An N-dimensional array.
    axis : int
        The axis along which the partial derivative is evaluated.
    dx : float, optional
        Scalar specifying the distance between sampling points along ``axis``.
    method : {'central', 'forward', 'backward'}, optional
        Finite difference method which is used in the interior of the domain
         of ``f``.
    out : `numpy.ndarray`, optional
         An N-dimensional array to which the output is written. Has to have
         the same shape as the input array ``f``.
    pad_mode : string, optional
        The padding mode to use outside the domain.

        ``'constant'``: Fill with ``pad_const``.

        ``'symmetric'``: Reflect at the boundaries, not doubling the
        outmost values.

        ``'periodic'``: Fill in values from the other side, keeping
        the order.

        ``'order0'``: Extend constantly with the outmost values
        (ensures continuity).

        ``'order1'``: Extend with constant slope (ensures continuity of
        the first derivative). This requires at least 2 values along
        each axis where padding is applied.

        ``'order2'``: Extend with second order accuracy (ensures continuity
        of the second derivative). This requires at least 3 values along
        each axis where padding is applied.

    pad_const : float, optional
        For ``pad_mode == 'constant'``, ``f`` assumes ``pad_const`` for
        indices outside the domain of ``f``

    Returns
    -------
    out : `numpy.ndarray`
        N-dimensional array of the same shape as ``f``. If ``out`` was
        provided, the returned object is a reference to it.

    Examples
    --------
    >>> f = np.array([ 0., 1., 2., 3., 4., 5., 6., 7., 8., 9.])

    >>> finite_diff(f, axis=0)
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  -9.])

    Without arguments the above defaults to:

    >>> finite_diff(f, axis=0, dx=1.0, method='forward', pad_mode='constant')
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  -9.])

    Parameters can be changed one by one:

    >>> finite_diff(f, axis=0, dx=0.5)
    array([ 2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  -18.])
    >>> finite_diff(f, axis=0, pad_mode='order1')
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1., 1.])

    Central differences and different edge orders:

    >>> finite_diff(0.5 * f ** 2, axis=0, method='central', pad_mode='order1')
    array([ 0.5,  1. ,  2. ,  3. ,  4. ,  5. ,  6. ,  7. ,  8. ,  8.5])
    >>> finite_diff(0.5 * f ** 2, axis=0, method='central', pad_mode='order2')
    array([-0.,  1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.,  9.])

    In-place evaluation:

    >>> out = f.copy()
    >>> out is finite_diff(f, axis=0, out=out)
    True
    """
    _, backend = get_array_and_backend(f)
    if out is None:
        return _finite_diff_pure(
            f, axis, dx=dx, method=method, pad_mode=pad_mode, pad_const=pad_const)
    else:
        if out.shape != f.shape:
            raise ValueError(f"expected output shape {f.shape}, got {out.shape}")
        out[:] = _finite_diff_pure(
            f, axis, dx=dx, method=method, pad_mode=pad_mode, pad_const=pad_const)
        return out



if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



"""Example of how to implement and use functionals."""

from __future__ import division, print_function
import odl


# Here we define the functional
class MyFunctional(odl.solvers.Functional):

    """This is my functional: ``||x||_2^2 + <x, y>``."""

    def __init__(self, space, y):
        """Initialize a new instance."""
        # This comand calls the init of Functional and sets a number of
        # parameters associated with a functional. All but domain have default
        # values if not set.
        super(MyFunctional, self).__init__(
            space=space, linear=False, grad_lipschitz=2)

        # We need to check that linear_term is in the domain. Then we store the
        # value of linear_term for future use.
        if y not in space:
            raise TypeError('linear_term is not in the domain!')
        self.y = y

    # Defining the _call function. This method is used for evaluation of
    # the functional and always needs to be implemented.
    def _call(self, x):
        """Evaluate the functional."""
        return x.norm() ** 2 + x.inner(self.y)

    # Next we define the gradient. Note that this is a property.
    @property
    def gradient(self):
        """The gradient operator."""

        # First we store the functional in a variable
        functional = self

        # The class corresponding to the gradient operator.
        class MyGradientOperator(odl.Operator):

            """Class implementing the gradient operator."""

            def __init__(self):
                """Initialize a new instance."""
                super(MyGradientOperator, self).__init__(
                    domain=functional.domain, range=functional.domain)

            def _call(self, x):
                """Evaluate the gradient."""
                # Here we can access the store functional from a few lines
                # above
                return 2.0 * x + functional.y

        return MyGradientOperator()

    # Next we define the convex conjugate functional.
    @property
    def convex_conj(self):
        """The convex conjugate functional."""
        # This functional is implemented below.
        return MyFunctionalConjugate(space=self.domain, y=self.y)


# Here is the conjugate functional. Note that this is a separate class, in
# contrast to the gradient which was implemented as an inner class. One
# advantage with the inner class it that we don't have to pass as many
# parameters when initializing, on the other hand having separate classes
# normally improves readibility of the code. Both methods are use throughout
# the odl package.
class MyFunctionalConjugate(odl.solvers.Functional):

    """Conjugate functional to ``||x||_2^2 + <x,y>``.

    This funtional has the analytic expression

        ``f^*(x) = ||x-y||^2/4``.
    """

    def __init__(self, space, y):
        """initialize a new instance."""
        super(MyFunctionalConjugate, self).__init__(
            space=space, linear=False, grad_lipschitz=2)

        if y not in space:
            raise TypeError('y is not in the domain!')
        self.y = y

    def _call(self, x):
        """Evaluate the functional."""
        return (x - self.y).norm()**2 / 4.0


# Create a functional
space = odl.uniform_discr(0, 1, 3)
linear_term = space.element([1, -4, 7])
my_func = MyFunctional(space=space, y=linear_term)

# Now we evaluate the functional in a random point
point = odl.core.util.testutils.noise_element(space)
print('Value of the functional in a random point: {}'
      ''.format(my_func(point)))

# Now we use the steepest-decent solver and backtracking linesearch in order to
# find the minimum of the functional.

# Create a starting guess. Also used by the solver to update in-place.
x = space.one()

# Create the linesearch object
line_search = odl.solvers.BacktrackingLineSearch(my_func, max_num_iter=10)

# Call the solver
odl.solvers.steepest_descent(my_func, x, maxiter=10, line_search=line_search)

print('Expected value: {}'.format((-1.0 / 2) * linear_term))
print('Found value: {}'.format(x))

# Create the convex conjugate functional of a scaled and translated functional
scalar = 3.2
translation = space.one()
scal_trans_cc_func = (scalar * my_func).translated(translation).convex_conj

# Evaluating the new functional in the random point.
print('Value of the new functional in a random point: {}'
      ''.format(scal_trans_cc_func(point)))



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Scipy compatibility module"""

import warnings
from os import environ
if 'SCIPY_ARRAY_API' in environ and environ['SCIPY_ARRAY_API']=='1':
        pass
else:
    warnings.warn("The environment variable SCIPY_ARRAY_API must be set to 1."
                + " It should be by default when importing odl, but it seems that scipy was imported before odl."
                + " If not set, the array API support of scipy will be disabled,"
                + " meaning that function calls such as ``xlogy`` on GPU will error"
                + " and throw back pytorch Type errors."
                + " Please add the following lines before your first scipy import. \n"
                + 'from os import environ \n' \
                + 'environ["SCIPY_ARRAY_API"]=="1" \n ' \
                + '********End of Warning********',
              stacklevel=2)

import scipy
import numpy

__all__ = (
    'lambertw',
    'scipy_lambertw',
    'xlogy',
    )

def _helper(operation: str, x1, x2=None, out=None, namespace=scipy.special, **kwargs):
    """Proxy for calling a scipy function on the data underlying `x1` and
    (if provided) `x2` and wrapping the result as an element of the suitable space.

    Args:
        operation (str): Name of the function
        x1  (Tensor): _description_
        x2  (Tensor, optional): _description_. Defaults to None.
        out (Tensor, optional): Out argument for in-place operations. Defaults to None.
        namespace: scipy namespace to get the operation from. Defaults to scipy.special.

    Returns:
        Tensor
    """
    return x1.space._elementwise_num_operation(
        operation=operation, x1=x1, x2=x2, out=out, namespace=namespace, **kwargs)

def lambertw(x, k=0, tol=1e-8):
    """
    Lambert W function.

    The Lambert W function W(z) is defined as the inverse function of w * exp(w). In other words, the value of W(z) is such that z = W(z) * exp(W(z)) for any complex number z.

    The Lambert W function is a multivalued function with infinitely many branches. Each branch gives a separate solution of the equation z = w exp(w). Here, the branches are indexed by the integer k.
    See https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.lambertw.html#scipy.special.lambertw
    """
    return _helper('lambertw', x, k=k, tol=tol)


def scipy_lambertw(x, k=0, tol=1e-8):
    """
    Lambert W function.

    The Lambert W function W(z) is defined as the inverse function of w * exp(w). In other words, the value of W(z) is such that z = W(z) * exp(W(z)) for any complex number z.

    The Lambert W function is a multivalued function with infinitely many branches. Each branch gives a separate solution of the equation z = w exp(w). Here, the branches are indexed by the integer k.
    See https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.lambertw.html#scipy.special.lambertw

    Note:
        This function is a direct call to scipy.special.lambertw on a Numpy Array!
    """
    assert isinstance(
        x, numpy.ndarray
    ), ("Can only call scipy_lambertw on nd_array. For ODL Tensors, please use the function scipy_compatibility.lambertw"
        + f". Got {type(x)}")
    return scipy.special.lambertw(x, k, tol)


def xlogy(x1, x2, out=None):
    """Compute x*log(y) so that the result is 0 if x = 0.
    See https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.xlogy.html
    """
    return _helper('xlogy', x1=x1, x2=x2, out=out)



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test for the figures of merit (FOMs) that use a known ground truth."""

from __future__ import division

import numpy as np
import pytest
import scipy.misc
import scipy.signal

import odl
from odl.core.set.space import LinearSpaceElement
from odl.contrib import fom
from odl.core.util.testutils import noise_element, simple_fixture, skip_if_no_pyfftw


# --- pytest fixtures --- #


fft_impl = simple_fixture(
    'fft_impl',
    [pytest.param('numpy'), pytest.param('pyfftw', marks=skip_if_no_pyfftw)]
)

space_ids =    [ "ℝ³ (float32)", "ℝ³ (float64)"
               , "uniform([0,1])", "uniform([0,1]²)" ]
space_params = [ ('ℝ³', 'float32'), ('ℝ³', 'float64')
               , ('uniform', 1), ('uniform', 2) ]

@pytest.fixture(scope="module", ids=space_ids, params=space_params)
def space(request, odl_impl_device_pairs):
    name, variation = request.param
    impl, device = odl_impl_device_pairs
    if name == 'ℝ³':
        dtype = variation
        return odl.rn(3, dtype=dtype, impl=impl, device=device)
    elif name == 'uniform':
        dimension = variation
        if dimension == 1:
            return odl.uniform_discr(0, 1, 10, impl=impl, device=device)
        elif dimension == 2:
            return odl.uniform_discr([0, 0], [1, 1], [5, 5], impl=impl, device=device)
     
    raise ValueError('undefined space')


scalar_fom = simple_fixture(
    'scalar_fom',
    [fom.mean_squared_error,
     fom.mean_absolute_error,
     fom.mean_value_difference,
     fom.standard_deviation_difference,
     fom.range_difference,
     fom.blurring]
)


# --- Tests --- #


def test_general(space, scalar_fom):
    """Test general properties of FOMs."""
    for name, ground_truth in space.examples:
        ground_truth = odl.abs(ground_truth)
        scale = odl.max(ground_truth) - odl.min(ground_truth)
        noise = scale * odl.abs(noise_element(space))
        data = ground_truth + noise

        # Check that range is a real number
        assert np.isscalar(scalar_fom(data, ground_truth))

        # Check that FOM is minimal when ground truth is compared with itself
        assert (
            scalar_fom(ground_truth, ground_truth)
            <= scalar_fom(data, ground_truth)
        )

        # Check that FOM is monotonic wrt noise level
        # This does not work for the FOMS `standard_deviation_difference`
        # and `range_difference`.
        if scalar_fom not in [fom.standard_deviation_difference,
                              fom.range_difference]:
            assert (
                scalar_fom(ground_truth + noise, ground_truth)
                <= scalar_fom(ground_truth + 2 * noise, ground_truth)
            )


def filter_image(image, fh, fv):
    """Reference filtering function using ``scipy.signal.convolve``."""
    fh, fv = np.asarray(fh), np.asarray(fv)
    image = scipy.signal.convolve(image, fh[:, None], mode='same')
    return scipy.signal.convolve(image, fv[None, :], mode='same')


def test_filter_image_fft(fft_impl):
    """Test the FFT filtering function against the real-space variant."""
    image = np.random.rand(128, 128)
    image -= np.mean(image)

    for fh, fv in zip([[1, 1], [1, 1, 1], [1, 1, 1, 1]],
                      [[-1, 1], [-1, 1, 1], [-1, -1, 1, 1]]):

        conv_real = filter_image(image, fh, fv)
        conv_fft = fom.util.filter_image_sep2d(image, fh, fv, impl=fft_impl)
        assert np.allclose(conv_real, conv_fft)


def test_mean_squared_error(space):
    true = odl.core.phantom.white_noise(space)
    data = odl.core.phantom.white_noise(space)

    result = fom.mean_squared_error(data, true)
    expected = odl.mean((true - data) ** 2)

    assert result == pytest.approx(expected)


def test_mean_absolute_error(space):
    true = odl.core.phantom.white_noise(space)
    data = odl.core.phantom.white_noise(space)

    result = fom.mean_absolute_error(data, true)
    expected = odl.mean(odl.abs(true - data))

    assert result == pytest.approx(expected)


def test_psnr(space):
    """Test the ``psnr`` fom."""
    true = odl.core.phantom.white_noise(space)
    data = odl.core.phantom.white_noise(space)
    zero = space.zero()

    # Check the corner cases
    assert fom.psnr(true, true) == np.inf
    assert fom.psnr(zero, zero) == np.inf
    assert fom.psnr(data, zero) == -np.inf

    # Compute the true value
    mse = odl.mean((true - data) ** 2)
    maxi = odl.max(odl.abs(true))
    expected = 10 * np.log10(maxi ** 2 / mse)

    # Test regular call
    result = fom.psnr(data, true)
    assert result == pytest.approx(expected, abs=1e-6)

    # Test with force_lower_is_better giving negative of expected
    result = fom.psnr(data, true, force_lower_is_better=True)
    assert result == pytest.approx(-expected, abs=1e-6)

    # Test with Z-score that result is independent of affine transformation
    result = fom.psnr(data * 3.7 + 1.234, true, use_zscore=True)
    expected = fom.psnr(data, true, use_zscore=True)
    assert result == pytest.approx(expected, abs=1e-5)


def test_ssim(space):
    ground_truth = odl.core.phantom.white_noise(space)

    # SSIM of true image should be either
    # * 1 with force_lower_is_better == False,
    # * -1 with force_lower_is_better == True and normalized == False,
    # * 0 with force_lower_is_better == True and normalized == True.
    result = fom.ssim(ground_truth, ground_truth)
    assert result == pytest.approx(1)

    result_normalized = fom.ssim(ground_truth, ground_truth, normalized=True)
    assert result_normalized == pytest.approx(1)

    result_flib = fom.ssim(ground_truth, ground_truth,
                           force_lower_is_better=True)
    assert result_flib == pytest.approx(-1)

    result_nf = fom.ssim(ground_truth, ground_truth, normalized=True,
                         force_lower_is_better=True)
    assert result_nf == pytest.approx(0)

    # SSIM with ground truth zero should always give zero if not normalized
    # and 1/2 otherwise.
    data = odl.core.phantom.white_noise(space)

    result = fom.ssim(data, space.zero())
    assert result == pytest.approx(0)

    result_normalized = fom.ssim(data, space.zero(), normalized=True)
    assert result_normalized == pytest.approx(0.5)

    result_flib = fom.ssim(data, space.zero(), force_lower_is_better=True)
    assert result_flib == pytest.approx(0)

    result_nf = fom.ssim(data, space.zero(), normalized=True,
                         force_lower_is_better=True)
    assert result_nf == pytest.approx(0.5)

    # SSIM should be symmetric if the dynamic range is set explicitly.
    for nor in [True, False]:
        for flib in [True, False]:
            result1 = fom.ssim(data, ground_truth, dynamic_range=1,
                               normalized=nor, force_lower_is_better=flib)
            result2 = fom.ssim(ground_truth, data, dynamic_range=1,
                               normalized=nor, force_lower_is_better=flib)
            assert result1 == pytest.approx(result2)


def test_mean_value_difference_sign():
    space = odl.uniform_discr(0, 1, 10)
    I0 = space.one()
    assert np.abs(fom.mean_value_difference(I0, -I0)) == pytest.approx(2.0)


def test_mean_value_difference_range_value(space):
    I0 = odl.core.util.testutils.noise_element(space)
    I1 = odl.core.util.testutils.noise_element(space)
    max0 = odl.max(I0)
    max1 = odl.max(I1)
    min0 = odl.min(I0)
    min1 = odl.min(I1)

    assert fom.mean_value_difference(I0, I1) <= max(max0 - min1, max1 - min0)
    assert fom.mean_value_difference(I0, I0) == pytest.approx(0)
    assert fom.mean_value_difference(10 * I0, I0, normalized=True) <= 1.0


def test_standard_deviation_difference_range_value(space):
    I0 = odl.core.util.testutils.noise_element(space)
    value_shift = np.random.normal(0, 10)

    assert fom.standard_deviation_difference(I0, I0) == pytest.approx(0)
    assert (
        fom.standard_deviation_difference(10 * I0, I0, normalized=True)
        <= 1.0
    )
    assert (
        fom.standard_deviation_difference(I0, I0 + value_shift)
        == pytest.approx(0, abs=1e-5)
    )
    test_value = fom.standard_deviation_difference(space.one(), space.zero(),
                                                   normalized=True)
    assert test_value == pytest.approx(0, abs=1e-6)


def test_range_difference(space):
    I0 = space.element(np.random.normal(0, 1, size=space.shape))
    I1 = space.element(np.random.normal(0, 1, size=space.shape))
    const = np.random.normal(0, 10)

    assert fom.range_difference(I0, I0) == pytest.approx(0)
    assert fom.range_difference(I0 + const, I0) == pytest.approx(0, abs=1e-5)
    aconst = np.abs(const)
    eval0 = aconst * fom.range_difference(I0, I1)
    eval1 = fom.range_difference(aconst * I0, aconst * I1)
    assert eval0 == pytest.approx(eval1, abs=1e-5)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Backend for ASTRA using CUDA."""


import warnings
from multiprocessing import Lock

import numpy as np
from packaging.version import parse as parse_version

from odl.core.discr import DiscretizedSpace
from odl.applications.tomo.backends.astra_setup import (
    ASTRA_VERSION, astra_projection_geometry,
    astra_projector, astra_supports, astra_versions_supporting,
    astra_volume_geometry)
from odl.applications.tomo.backends.util import _add_default_complex_impl
from odl.applications.tomo.geometry import (
    ConeBeamGeometry, FanBeamGeometry, Geometry, Parallel2dGeometry,
    Parallel3dAxisGeometry)
from odl.core.discr.discr_space import DiscretizedSpaceElement
from odl.core.array_API_support import empty, get_array_and_backend

try:
    import astra

    # This is important, although not use explicitely.
    # If not imported, astra.experimental is not "visible"
    import astra.experimental

    ASTRA_CUDA_AVAILABLE = astra.astra.use_cuda()
except ImportError:
    ASTRA_CUDA_AVAILABLE = False

__all__ = (
    'ASTRA_CUDA_AVAILABLE',
)
   

def index_of_cuda_device(device: "torch.device"):
    """Gets the integer index of a cuda device"""
    if device == 'cpu':
        return None
    else:
        return int(str(device).split(':')[-1])

class AstraCudaImpl:
    """`RayTransform` implementation for CUDA algorithms in ASTRA."""

    projector_id = None

    def __init__(self, geometry, vol_space, proj_space):
        """Initialize a new instance.

        Parameters
        ----------
        geometry : `Geometry`
            Geometry defining the tomographic setup.
        vol_space : `DiscretizedSpace`
            Reconstruction space, the space of the images to be forward
            projected.
        proj_space : `DiscretizedSpace`
            Projection space, the space of the result.
        """
        if not isinstance(geometry, Geometry):
            raise TypeError(f"`geometry` must be a `Geometry` instance, got {geometry}")
        if not isinstance(vol_space, DiscretizedSpace):
            raise TypeError(
                f"`vol_space` must be a `DiscretizedSpace` instance, got {vol_space}"
            )
        if not isinstance(proj_space, DiscretizedSpace):
            raise TypeError(
                f"`proj_space` must be a `DiscretizedSpace` instance, got {proj_space}"
            )

        # Print a warning if the detector midpoint normal vector at any
        # angle is perpendicular to the geometry axis in parallel 3d
        # single-axis geometry -- this is broken in some ASTRA versions
        if (
            isinstance(geometry, Parallel3dAxisGeometry)
            and not astra_supports('par3d_det_mid_pt_perp_to_axis')
        ):
            req_ver = astra_versions_supporting('par3d_det_mid_pt_perp_to_axis')
            axis = geometry.axis
            mid_pt = geometry.det_params.mid_pt
            for i, angle in enumerate(geometry.angles):
                if abs(np.dot(axis, geometry.det_to_src(angle, mid_pt))) < 1e-4:
                    warnings.warn(
                        f"angle {i}: detector midpoint normal {geometry.det_to_src(angle, mid_pt)}"
                        + f" is perpendicular to the geometry axis {axis} in `Parallel3dAxisGeometry`;"
                        + f" this is broken in ASTRA {ASTRA_VERSION}, please upgrade to ASTRA {req_ver}",
                        RuntimeWarning,
                    )
                    break

        self.geometry = geometry
        self._vol_space = vol_space
        self._proj_space = proj_space

        self.create_ids()

        # ASTRA projectors are not thread-safe, thus we need to lock manually
        self._mutex = Lock()
        assert (
            vol_space.impl == proj_space.impl
        ), f"Volume space ({vol_space.impl}) != Projection space ({proj_space.impl})"

        if self.geometry.ndim == 3:
            if vol_space.impl == 'numpy':
                self.transpose_tuple = (1,0,2)
            elif vol_space.impl == 'pytorch':
                self.transpose_tuple = (1,0)
            else:
                raise NotImplementedError("Not implemented for another backend")

        self.fp_scaling_factor = astra_cuda_fp_scaling_factor(self.geometry)
        self.bp_scaling_factor = astra_cuda_bp_scaling_factor(
                             self.proj_space, self.vol_space, self.geometry)

    @property
    def vol_space(self):
        """Volume Space of the Ray Transform"""
        return self._vol_space

    @property
    def proj_space(self):
        """Projection Space of the Ray Transform"""
        return self._proj_space

    def create_ids(self):
        """Create ASTRA objects."""
        # Create input and output arrays
        if self.geometry.motion_partition.ndim == 1:
            motion_shape = self.geometry.motion_partition.shape
        else:
            # Need to flatten 2- or 3-dimensional angles into one axis
            motion_shape = (np.prod(self.geometry.motion_partition.shape),)

        proj_shape = motion_shape + self.geometry.det_partition.shape
        self.proj_ndim = len(proj_shape)

        # Create ASTRA data structures
        self.vol_geom  = astra_volume_geometry(self.vol_space, 'cuda')
        
        self.proj_geom = astra_projection_geometry(self.geometry, 'cuda')

        self.projector_id = astra_projector(
            astra_proj_type = 'cuda3d', 
            astra_vol_geom  = self.vol_geom, 
            astra_proj_geom = self.proj_geom, 
            ndim = 3, 
            override_2D = bool(self.geometry.ndim == 2)
        )        

    @_add_default_complex_impl
    def call_forward(self, x, out=None, **kwargs):
        """Run an ASTRA forward projection on the given data using the GPU.

        Parameters
        ----------
        vol_data : ``vol_space.real_space`` element
            Volume data to which the projector is applied. Although
            ``vol_space`` may be complex, this element needs to be real.
        out : ``proj_space`` element, optional
            Element of the projection space to which the result is written. If
            ``None``, an element in `proj_space` is created.

        Returns
        -------
        out : ``proj_space`` element
            Projection data resulting from the application of the projector.
            If ``out`` was provided, the returned object is a reference to it.
        """
        return self._call_forward_real(x, out, **kwargs)

    def _call_forward_real(self, vol_data:DiscretizedSpaceElement, out=None, **kwargs):
        """Run an ASTRA forward projection on the given data using the GPU.

        Parameters
        ----------
        vol_data : ``vol_space.real_space`` element
            Volume data to which the projector is applied. Although
            ``vol_space`` may be complex, this element needs to be real.
        out : ``proj_space`` element, optional
            Element of the projection space to which the result is written. If
            ``None``, an element in `proj_space` is created.

        Returns
        -------
        out : ``proj_space`` element
            Projection data resulting from the application of the projector.
            If ``out`` was provided, the returned object is a reference to it.
        """
        with self._mutex:
            assert vol_data in self.vol_space.real_space

            if out is not None:
                assert (
                    out in self.proj_space.real_space
                ), ("The out argument provided is a {type(out)}, which is not an element"
                    + f" of the projection space {self.proj_space.real_space}")
                if self.vol_space.impl == "pytorch":
                    warnings.warn(
                        "You requested an out-of-place transform with PyTorch."
                        + " This will require cloning the data and will allocate extra memory",
                        RuntimeWarning,
                    )
                proj_data = out.data[None] if self.proj_ndim == 2 else out.data
                if self.geometry.ndim == 3:
                    proj_data = proj_data.transpose(*self.transpose_tuple)

            else:
                proj_data = empty(
                    impl   = self.proj_space.impl,
                    shape  = astra.geom_size(self.proj_geom),
                    dtype  = self.proj_space.dtype,
                    device = self.proj_space.device
                )

            if self.proj_ndim == 2:
                volume_data = vol_data.data[None]
            elif self.proj_ndim == 3:
                volume_data = vol_data.data
            else:
                raise NotImplementedError

            volume_data, _ = get_array_and_backend(volume_data, must_be_contiguous=True)
            proj_data, _ = get_array_and_backend(proj_data, must_be_contiguous=True)

            if self.proj_space.impl == 'pytorch':
                device_index = index_of_cuda_device(
                                  self.proj_space.tspace.device) #type:ignore
                if device_index is not None:
                    astra.set_gpu_index(device_index)
                    
            astra.experimental.direct_FP3D( #type:ignore
                self.projector_id,
                volume_data,
                proj_data
            )

            proj_data *= self.fp_scaling_factor
            proj_data = (
                proj_data[0]
                if self.geometry.ndim == 2
                else proj_data.transpose(*self.transpose_tuple)
            )

            if out is not None:
                out.data[:] = proj_data if self.proj_space.impl == 'numpy' else proj_data.clone()
            else:
                return self.proj_space.element(proj_data)

    @_add_default_complex_impl
    def call_backward(self, x, out=None, **kwargs):
        """Run an ASTRA back-projection on the given data using the GPU.

        Parameters
        ----------
        proj_data : ``proj_space.real_space`` element
            Projection data to which the back-projector is applied. Although
            ``proj_space`` may be complex, this element needs to be real.
        out : ``vol_space`` element, optional
            Element of the reconstruction space to which the result is written.
            If ``None``, an element in ``vol_space`` is created.

        Returns
        -------
        out : ``vol_space`` element
            Reconstruction data resulting from the application of the
            back-projector. If ``out`` was provided, the returned object is a
            reference to it.
        """
        return self._call_backward_real(x, out, **kwargs)

    def _call_backward_real(self, proj_data:DiscretizedSpaceElement, out=None, **kwargs):
        """Run an ASTRA back-projection on the given data using the GPU.

        Parameters
        ----------
        proj_data : ``proj_space.real_space`` element
            Projection data to which the back-projector is applied. Although
            ``proj_space`` may be complex, this element needs to be real.
        out : ``vol_space`` element, optional
            Element of the reconstruction space to which the result is written.
            If ``None``, an element in ``vol_space`` is created.

        Returns
        -------
        out : ``vol_space`` element
            Reconstruction data resulting from the application of the
            back-projector. If ``out`` was provided, the returned object is a
            reference to it.
        """
        with self._mutex:
            assert proj_data in self.proj_space.real_space

            if out is not None:
                assert (
                    out in self.vol_space.real_space
                ), f"The out argument provided is a {type(out)}, which is not an element of the projection space {self.vol_space.real_space}"
                if self.vol_space.impl == 'pytorch':
                    warnings.warn(
                        "You requested an out-of-place transform with PyTorch. \
                        This will require cloning the data and will allocate extra memory", 
                        RuntimeWarning)
                volume_data = out.data[None] if self.geometry.ndim==2 else out.data
            else:
                volume_data = empty(
                    self.vol_space.impl,
                    astra.geom_size(self.vol_geom),
                    dtype  = self.vol_space.dtype,
                    device = self.vol_space.device
                )

            ### Transpose projection tensor
            if self.proj_ndim == 2:
                proj_data = proj_data.data[None]
            elif self.proj_ndim == 3:
                proj_data = proj_data.data.transpose(*self.transpose_tuple)
            else:
                raise NotImplementedError

            # Ensure data is contiguous otherwise astra will throw an error
            volume_data, _ = get_array_and_backend(volume_data, must_be_contiguous=True)
            proj_data, _ = get_array_and_backend(proj_data, must_be_contiguous=True)
            
            if self.vol_space.tspace.impl == 'pytorch':
                device_index = index_of_cuda_device(self.vol_space.tspace.device) #type:ignore
                if device_index is not None:
                    astra.set_gpu_index(device_index)

            ### Call the backprojection
            astra.experimental.direct_BP3D( #type:ignore
                self.projector_id,
                volume_data,
                proj_data                
            )
            volume_data *= self.bp_scaling_factor
            volume_data = volume_data[0] if self.geometry.ndim == 2 else volume_data

            if out is not None:
                out[:] = volume_data if self.vol_space.impl == 'numpy' else volume_data.clone()
                return out
            else:
                return self.vol_space.element(volume_data)


def astra_cuda_fp_scaling_factor(geometry):
    """Volume scaling accounting for differing adjoint definitions.

    ASTRA defines the adjoint operator in terms of a fully discrete
    setting (transposed "projection matrix") without any relation to
    physical dimensions, which makes a re-scaling necessary to
    translate it to spaces with physical dimensions.

    Behavior of ASTRA changes slightly between versions, so we keep
    track of it and adapt the scaling accordingly.
    """
    if (
        isinstance(geometry, Parallel2dGeometry)
        and parse_version(ASTRA_VERSION) < parse_version('1.9.9.dev')
    ):
        # parallel2d scales with pixel stride
        return 1 / float(geometry.det_partition.cell_volume)
    
    else:
        return 1

def astra_cuda_bp_scaling_factor(proj_space, vol_space, geometry):
    """Volume scaling accounting for differing adjoint definitions.

    ASTRA defines the adjoint operator in terms of a fully discrete
    setting (transposed "projection matrix") without any relation to
    physical dimensions, which makes a re-scaling necessary to
    translate it to spaces with physical dimensions.

    Behavior of ASTRA changes slightly between versions, so we keep
    track of it and adapt the scaling accordingly.
    """
    # Angular integration weighting factor
    # angle interval weight by approximate cell volume
    angle_extent = geometry.motion_partition.extent
    num_angles = geometry.motion_partition.shape
    # TODO: this gives the wrong factor for Parallel3dEulerGeometry with
    # 2 angles
    scaling_factor = (angle_extent / num_angles).prod()

    # Correct in case of non-weighted spaces
    proj_extent = float(proj_space.partition.extent.prod())
    proj_size = float(proj_space.partition.size)
    proj_weighting = proj_extent / proj_size

    scaling_factor *= proj_space.weighting.const / proj_weighting
    scaling_factor /= vol_space.weighting.const / vol_space.cell_volume

    if parse_version(ASTRA_VERSION) < parse_version('1.8rc1'):
        # Scaling for the old, pre-1.8 behaviour
        if isinstance(geometry, Parallel2dGeometry):
            # Scales with 1 / cell_volume
            scaling_factor *= float(vol_space.cell_volume)
        elif (isinstance(geometry, FanBeamGeometry)
              and geometry.det_curvature_radius is None):
            # Scales with 1 / cell_volume
            scaling_factor *= float(vol_space.cell_volume)
            # Additional magnification correction
            src_radius = geometry.src_radius
            det_radius = geometry.det_radius
            scaling_factor *= ((src_radius + det_radius) / src_radius)
        elif isinstance(geometry, Parallel3dAxisGeometry):
            # Scales with voxel stride
            # In 1.7, only cubic voxels are supported
            voxel_stride = vol_space.cell_sides[0]
            scaling_factor /= float(voxel_stride)
        elif (isinstance(geometry, ConeBeamGeometry)
              and geometry.det_curvature_radius is None):
            # Scales with 1 / cell_volume
            # In 1.7, only cubic voxels are supported
            voxel_stride = vol_space.cell_sides[0]
            scaling_factor /= float(voxel_stride)
            # Magnification correction
            src_radius = geometry.src_radius
            det_radius = geometry.det_radius
            scaling_factor *= ((src_radius + det_radius) / src_radius) ** 2
    elif parse_version(ASTRA_VERSION) < parse_version('1.9.0dev'):
        # Scaling for the 1.8.x releases
        if isinstance(geometry, Parallel2dGeometry):
            # Scales with 1 / cell_volume
            scaling_factor *= float(vol_space.cell_volume)
        elif (isinstance(geometry, FanBeamGeometry)
              and geometry.det_curvature_radius is None):
            # Scales with 1 / cell_volume
            scaling_factor *= float(vol_space.cell_volume)
            # Magnification correction
            src_radius = geometry.src_radius
            det_radius = geometry.det_radius
            scaling_factor *= ((src_radius + det_radius) / src_radius)
        elif isinstance(geometry, Parallel3dAxisGeometry):
            # Scales with cell volume
            # currently only square voxels are supported
            scaling_factor /= vol_space.cell_volume
        elif (isinstance(geometry, ConeBeamGeometry)
              and geometry.det_curvature_radius is None):
            # Scales with cell volume
            scaling_factor /= vol_space.cell_volume
            # Magnification correction (scaling = 1 / magnification ** 2)
            src_radius = geometry.src_radius
            det_radius = geometry.det_radius
            scaling_factor *= ((src_radius + det_radius) / src_radius) ** 2

            # Correction for scaled 1/r^2 factor in ASTRA's density weighting.
            # This compensates for scaled voxels and pixels, as well as a
            # missing factor src_radius ** 2 in the ASTRA BP with
            # density weighting.
            det_px_area = geometry.det_partition.cell_volume
            scaling_factor *= (
                src_radius ** 2 * det_px_area ** 2 / vol_space.cell_volume ** 2
            )
    elif parse_version(ASTRA_VERSION) < parse_version('1.9.9.dev'):
        # Scaling for intermediate dev releases between 1.8.3 and 1.9.9.dev
        if isinstance(geometry, Parallel2dGeometry):
            # Scales with 1 / cell_volume
            scaling_factor *= float(vol_space.cell_volume)
        elif (isinstance(geometry, FanBeamGeometry)
              and geometry.det_curvature_radius is None):
            # Scales with 1 / cell_volume
            scaling_factor *= float(vol_space.cell_volume)
            # Magnification correction
            src_radius = geometry.src_radius
            det_radius = geometry.det_radius
            scaling_factor *= ((src_radius + det_radius) / src_radius)
        elif isinstance(geometry, Parallel3dAxisGeometry):
            # Scales with cell volume
            # currently only square voxels are supported
            scaling_factor /= vol_space.cell_volume
        elif (isinstance(geometry, ConeBeamGeometry)
              and geometry.det_curvature_radius is None):
            # Scales with cell volume
            scaling_factor /= vol_space.cell_volume
            # Magnification correction (scaling = 1 / magnification ** 2)
            src_radius = geometry.src_radius
            det_radius = geometry.det_radius
            scaling_factor *= ((src_radius + det_radius) / src_radius) ** 2

            # Correction for scaled 1/r^2 factor in ASTRA's density weighting.
            # This compensates for scaled voxels and pixels, as well as a
            # missing factor src_radius ** 2 in the ASTRA BP with
            # density weighting.
            det_px_area = geometry.det_partition.cell_volume
            scaling_factor *= (src_radius ** 2 * det_px_area ** 2)
    else:
        # Scaling for versions since 1.9.9.dev
        scaling_factor /= float(vol_space.cell_volume)
        scaling_factor *= float(geometry.det_partition.cell_volume)

    return scaling_factor


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests

    run_doctests()



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""NumPy implementation of tensor spaces."""

import sys

import array_api_compat.numpy as xp

from odl.core.set.space import LinearSpaceElement
from odl.core.space.base_tensors import Tensor, TensorSpace
from odl.core.util import is_numeric_dtype
from odl.core.array_API_support import ArrayBackend, lookup_array_backend

__all__ = ('NumpyTensorSpace', 'numpy_array_backend')


def _npy_to_device(x:xp.ndarray, device : str) -> xp.ndarray:
    """Internal dummy function used to "change" the device of a np.ndarray.
    It will raise a ValueError for any device but 'cpu', as it is not
    actually possible to switch to any other device in NumPy. The purpose
    of this is only to confirm to a shared API with other backends that
    do support this.

    Args:
        x (xp.ndarray): array to change the device of
        device (str): string identifier of the target device

    Raises:
        ValueError: Raised for any device argument but 'cpu'

    Returns:
        (xp.ndarray): Array on the target device
    """
    if device == "cpu":
        return x
    else:
        raise ValueError(f"NumPy only supports device CPU, not {device}.")

try:
    numpy_array_backend = ArrayBackend(
        impl = 'numpy',
        available_dtypes = {
          key : xp.dtype(key) for key in [
            'bool',
            'int8',
            'int16',
            'int32',
            'int64',
            'uint8',
            'uint16',
            'uint32',
            'uint64',
            'float32',
            'float64',
            'complex64',
            'complex128',
          ]},
        array_namespace = xp,
        array_constructor = xp.asarray,
        from_dlpack = xp.from_dlpack,
        array_type = xp.ndarray,
        make_contiguous = lambda x: x if x.data.c_contiguous else xp.ascontiguousarray(x),
        identifier_of_dtype = lambda dt: xp.dtype(dt).name,
        available_devices = ['cpu'],
        to_cpu = lambda x: x,
        to_numpy = lambda x : x,
        to_device = _npy_to_device
     )
except KeyError:
    # PyTest runs modules twice, causing a "duplicate" registration of the backend.
    # Otherwise this should not happen.
    assert "pytest" in sys.modules
    numpy_array_backend = lookup_array_backend('numpy')

class NumpyTensorSpace(TensorSpace):

    """Set of tensors of arbitrary data type, implemented with NumPy.

    A tensor is, in the most general sense, a multi-dimensional array
    that allows operations per entry (keep the rank constant),
    reductions / contractions (reduce the rank) and broadcasting
    (raises the rank).
    For non-numeric data type like ``object``, the range of valid
    operations is rather limited since such a set of tensors does not
    define a vector space.
    Any numeric data type, on the other hand, is considered valid for
    a tensor space, although certain operations - like division with
    integer dtype - are not guaranteed to yield reasonable results.

    Under these restrictions, all basic vector space operations are
    supported by this class, along with reductions based on arithmetic
    or comparison, and element-wise mathematical functions ("ufuncs").

    This class is implemented using `numpy.ndarray`'s as back-end.

    See the `Wikipedia article on tensors`_ for further details.
    See also [Hac2012] "Part I Algebraic Tensors" for a rigorous
    treatment of tensors with a definition close to this one.

    Note also that this notion of tensors is the same as in popular
    Deep Learning frameworks.

    References
    ----------
    [Hac2012] Hackbusch, W. *Tensor Spaces and Numerical Tensor Calculus*.
    Springer, 2012.

    .. _Wikipedia article on tensors: https://en.wikipedia.org/wiki/Tensor
    """

    def __init__(self, shape, dtype='float64', device = 'cpu', **kwargs):
        r"""Initialize a new instance.

        Parameters
        ----------
        shape : positive int or sequence of positive ints
            Number of entries per axis for elements in this space. A
            single integer results in a space with rank 1, i.e., 1 axis.
        dtype (str): optional
            Data type of each element. Defaults to 'float64'
        device (str):
            Device on which the data is. For Numpy, it must be 'cpu'.

        Other Parameters
        ----------------
        weighting : optional
            Use weighted inner product, norm, and dist. The following
            types are supported as ``weighting``:

            ``None``: no weighting, i.e. weighting with ``1.0`` (default).

            `Weighting`: Use this weighting as-is. Compatibility
            with this space's elements is not checked during init.

            ``float``: Weighting by a constant.

            array-like: Pointwise weighting by an array.

            This option cannot be combined with ``dist``,
            ``norm`` or ``inner``. It also cannot be used in case of
            non-numeric ``dtype``.

        dist : callable, optional
            Distance function defining a metric on the space.
            It must accept two `NumpyTensor` arguments and return
            a non-negative real number. See ``Notes`` for
            mathematical requirements.

            By default, ``dist(x, y)`` is calculated as ``norm(x - y)``.

            This option cannot be combined with ``weight``,
            ``norm`` or ``inner``. It also cannot be used in case of
            non-numeric ``dtype``.

        norm : callable, optional
            The norm implementation. It must accept a
            `NumpyTensor` argument, return a non-negative real number.
            See ``Notes`` for mathematical requirements.

            By default, ``norm(x)`` is calculated as ``inner(x, x)``.

            This option cannot be combined with ``weight``,
            ``dist`` or ``inner``. It also cannot be used in case of
            non-numeric ``dtype``.

        inner : callable, optional
            The inner product implementation. It must accept two
            `NumpyTensor` arguments and return an element of the field
            of the space (usually real or complex number).
            See ``Notes`` for mathematical requirements.

            This option cannot be combined with ``weight``,
            ``dist`` or ``norm``. It also cannot be used in case of
            non-numeric ``dtype``.

        exponent : positive float, optional
            Exponent of the norm. For values other than 2.0, no
            inner product is defined.

            This option has no impact if either ``dist``, ``norm`` or
            ``inner`` is given, or if ``dtype`` is non-numeric.

            Default: 2.0

        kwargs :
            Further keyword arguments are passed to the weighting
            classes.

        See Also
        --------
        odl.core.space.space_utils.rn : constructor for real tensor spaces
        odl.core.space.space_utils.cn : constructor for complex tensor spaces
        odl.core.space.space_utils.tensor_space :
            constructor for tensor spaces of arbitrary scalar data type

        Notes
        -----
        - A distance function or metric on a space :math:`\mathcal{X}`
          is a mapping
          :math:`d:\mathcal{X} \times \mathcal{X} \to \mathbb{R}`
          satisfying the following conditions for all space elements
          :math:`x, y, z`:

          * :math:`d(x, y) \geq 0`,
          * :math:`d(x, y) = 0 \Leftrightarrow x = y`,
          * :math:`d(x, y) = d(y, x)`,
          * :math:`d(x, y) \leq d(x, z) + d(z, y)`.

        - A norm on a space :math:`\mathcal{X}` is a mapping
          :math:`\| \cdot \|:\mathcal{X} \to \mathbb{R}`
          satisfying the following conditions for all
          space elements :math:`x, y`: and scalars :math:`s`:

          * :math:`\| x\| \geq 0`,
          * :math:`\| x\| = 0 \Leftrightarrow x = 0`,
          * :math:`\| sx\| = |s| \cdot \| x \|`,
          * :math:`\| x+y\| \leq \| x\| +
            \| y\|`.

        - An inner product on a space :math:`\mathcal{X}` over a field
          :math:`\mathbb{F} = \mathbb{R}` or :math:`\mathbb{C}` is a
          mapping
          :math:`\langle\cdot, \cdot\rangle: \mathcal{X} \times
          \mathcal{X} \to \mathbb{F}`
          satisfying the following conditions for all
          space elements :math:`x, y, z`: and scalars :math:`s`:

          * :math:`\langle x, y\rangle =
            \overline{\langle y, x\rangle}`,
          * :math:`\langle sx + y, z\rangle = s \langle x, z\rangle +
            \langle y, z\rangle`,
          * :math:`\langle x, x\rangle = 0 \Leftrightarrow x = 0`.

        Examples
        --------
        Explicit initialization with the class constructor:

        >>> space = NumpyTensorSpace(3, float)
        >>> space
        rn(3)
        >>> space.shape
        (3,)
        >>> space.dtype
        dtype('float64')

        A more convenient way is to use factory functions:

        >>> space = odl.rn(3, weighting=[1, 2, 3])
        >>> space
        rn(3, weighting=[1, 2, 3])
        >>> space = odl.tensor_space((2, 3), dtype=int)
        >>> space
        tensor_space((2, 3), 'int32')
        """
        super().__init__(shape, dtype, device, **kwargs)

    ########## Attributes ##########
    @property
    def array_backend(self) -> ArrayBackend:
        """
        ArrayBackend Object for the TensorSpace, here ``numpy_array_backend``
        """
        return numpy_array_backend

    @property
    def array_namespace(self):
        """Name of the array_namespace"""
        return xp

    @property
    def element_type(self):
        """Type of elements in this space: `NumpyTensor`."""
        return NumpyTensor

    @property
    def impl(self):
        """Name of the implementation back-end: ``'numpy'``."""
        return 'numpy'

    ######### public methods #########
    def broadcast_to(self, inp: int|float|complex|list|tuple|xp.ndarray) -> xp.ndarray:
        """Broadcasts the an inputs to the shape of the TensorSpace.
        Also performs a device conversion if necessary

        Args:
            inp (int | float | complex | list | tuple | xp.ndarray): Input to be broadcasted

        Returns:
            xp.ndarray: Returned array broadcasted to the TensorSpace shape and device
        """
        arr = self.array_namespace.broadcast_to(
                    self.array_namespace.asarray(inp, device=self.device),
                    self.shape
                    )
        # Make sure the result is writeable, if not make copy.
        # This happens for e.g. results of `np.broadcast_to()`.
        if not arr.flags.writeable:
            arr = arr.copy()
        return arr

    ######### private methods #########

class NumpyTensor(Tensor):

    """Representation of a `NumpyTensorSpace` element.

    This is an internal ODL class; it should not directly be instantiated from user code.
    Instead, always use the `.element` method of the `space` for generating these objects."""
    
    def __init__(self, space, data):
        """Initialize a new instance. The input must be a NumPy array."""
        LinearSpaceElement.__init__(self, space)
        assert(isinstance(data, xp.ndarray)), f"{type(data)=}, should be np.ndarray"
        if data.dtype != space.dtype:
            data = data.astype(space.dtype)
        self.__data = data

    @property
    def data(self):
        """The `numpy.ndarray` representing the data of ``self``."""
        return self.__data

    @data.setter
    def data(self, value):
        self.__data = value

    def _assign(self, other, avoid_deep_copy):
        """Assign the values of ``other``, which is assumed to be in the
        same space, to ``self``."""
        if avoid_deep_copy:
            self.__data = other.__data
        else:
            self.__data[:] = other.__data

    ######### Public methods #########
    def copy(self):
        """Return an identical (deep) copy of this tensor.

        Parameters
        ----------
        None

        Returns
        -------
        copy : `NumpyTensor`
            The deep copy

        Examples
        --------
        >>> space = odl.rn(3)
        >>> x = space.element([1, 2, 3])
        >>> y = x.copy()
        >>> y == x
        True
        >>> y is x
        False
        """
        return self.space.element(self.data.copy())

    def __getitem__(self, indices):
        """Return ``self[indices]``.

        Parameters
        ----------
        indices : index expression
            Integer, slice or sequence of these, defining the positions
            of the data array which should be accessed.

        Returns
        -------
        values : `NumpyTensorSpace.dtype` or `NumpyTensor`
            The value(s) at the given indices. Note that the returned
            object is a writable view into the original tensor, except
            for the case when ``indices`` is a list.

        Examples
        --------
        For one-dimensional spaces, indexing is as in linear arrays:

        >>> space = odl.rn(3)
        >>> x = space.element([1, 2, 3])
        >>> x[0]
        1.0
        >>> x[1:]
        rn(2).element([ 2.,  3.])

        In higher dimensions, the i-th index expression accesses the
        i-th axis:

        >>> space = odl.rn((2, 3))
        >>> x = space.element([[1, 2, 3],
        ...                    [4, 5, 6]])
        >>> x[0, 1]
        2.0
        >>> x[:, 1:]
        rn((2, 2)).element(
            [[ 2.,  3.],
             [ 5.,  6.]]
        )

        Slices can be assigned to, except if lists are used for indexing:

        >>> y = x[:, ::2]  # view into x
        >>> y[:] = -9
        >>> x
        rn((2, 3)).element(
            [[-9.,  2., -9.],
             [-9.,  5., -9.]]
        )
        >>> y = x[[0, 1], [1, 2]]  # not a view, won't modify x
        >>> y
        rn(2).element([ 2., -9.])
        >>> y[:] = 0
        >>> x
        rn((2, 3)).element(
            [[-9.,  2., -9.],
             [-9.,  5., -9.]]
        )
        """
        # Lazy implementation: index the array and deal with it
        if isinstance(indices, NumpyTensor):
            indices = indices.data
        arr = self.data[indices]

        if xp.isscalar(arr):
            if self.space.field is not None:
                return self.space.field.element(arr)
            else:
                return arr
        else:
            if is_numeric_dtype(self.dtype):
                weighting = self.space.weighting
            else:
                weighting = None
            space = type(self.space)(
                arr.shape, dtype=self.dtype, exponent=self.space.exponent,
                weighting=weighting)
            return space.element(arr, copy=False)

    def __setitem__(self, indices, values):
        """Implement ``self[indices] = values``.

        Parameters
        ----------
        indices : index expression
            Integer, slice or sequence of these, defining the positions
            of the data array which should be written to.
        values : scalar, array-like or `NumpyTensor`
            The value(s) that are to be assigned.

            If ``index`` is an integer, ``value`` must be a scalar.

            If ``index`` is a slice or a sequence of slices, ``value``
            must be broadcastable to the shape of the slice.

        Examples
        --------
        For 1d spaces, entries can be set with scalars or sequences of
        correct shape:

        >>> space = odl.rn(3)
        >>> x = space.element([1, 2, 3])
        >>> x[0] = -1
        >>> x[1:] = (0, 1)
        >>> x
        rn(3).element([-1.,  0.,  1.])

        It is also possible to use tensors of other spaces for
        casting and assignment:

        >>> space = odl.rn((2, 3))
        >>> x = space.element([[1, 2, 3],
        ...                    [4, 5, 6]])
        >>> x[0, 1] = -1
        >>> x
        rn((2, 3)).element(
            [[ 1., -1.,  3.],
             [ 4.,  5.,  6.]]
        )
        >>> short_space = odl.tensor_space((2, 2), dtype='int32')
        >>> y = short_space.element([[-1, 2],
        ...                          [0, 0]])
        >>> x[:, :2] = y
        >>> x
        rn((2, 3)).element(
            [[-1.,  2.,  3.],
             [ 0.,  0.,  6.]]
        )

        The Numpy assignment and broadcasting rules apply:

        >>> x[:] = np.array([[0, 0, 0],
        ...                  [1, 1, 1]])
        >>> x
        rn((2, 3)).element(
            [[ 0.,  0.,  0.],
             [ 1.,  1.,  1.]]
        )
        >>> x[:, 1:] = [7, 8]
        >>> x
        rn((2, 3)).element(
            [[ 0.,  7.,  8.],
             [ 1.,  7.,  8.]]
        )
        >>> x[:, ::2] = -2.
        >>> x
        rn((2, 3)).element(
            [[-2.,  7., -2.],
             [-2.,  7., -2.]]
        )
        """
        if isinstance(indices, type(self)):
            indices = indices.data
        if isinstance(values, type(self)):
            values = values.data

        self.data[indices] = values

if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

# pylint: disable=line-too-long

"""Scipy implementation of Sparse Matrices."""

from scipy.sparse import coo_matrix

from odl.core.sparse.sparse_template import SparseMatrixFormat, _registered_sparse_formats

if ('scipy' not in _registered_sparse_formats
        or 'COO' not in _registered_sparse_formats['scipy']):
    scipy_coo_tensor = SparseMatrixFormat(
        sparse_format='COO',
        impl = 'scipy',
        constructor = coo_matrix,
        is_of_this_sparse_format = lambda x : isinstance(x, coo_matrix),
        to_dense = lambda matrix: matrix.toarray(),
        matmul_spmatrix_with_vector = lambda matrix, x: matrix.dot(x)
    )




# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test configuration file."""

from __future__ import absolute_import, division, print_function

import operator
import os
from os import path

import numpy as np

import odl
from odl.core.array_API_support import lookup_array_backend, check_device
from odl.core.space.entry_points import tensor_space_impl_names
from odl.trafos.backends import PYFFTW_AVAILABLE, PYWT_AVAILABLE
from odl.core.util.testutils import simple_fixture
from odl.core.util.dtype_utils import INTEGER_DTYPES, FLOAT_DTYPES, COMPLEX_DTYPES

try:
    import pytest
    from pytest import fixture
except ImportError:
    pytest = None

    # Identity fixture
    def fixture(*arg, **kw):
        if arg and callable(arg[0]):
            return arg[0]
        return fixture


# --- Add numpy and ODL to all doctests ---


@fixture(autouse=True)
def _add_doctest_np_odl(doctest_namespace):
    doctest_namespace['np'] = np
    doctest_namespace['odl'] = odl


# --- Ignored tests due to missing modules ---


this_dir = path.dirname(__file__)
odl_root = path.abspath(path.join(this_dir, '..', '..'))
collect_ignore = [path.join(odl_root, 'setup.py'),
                  path.join(odl_root, 'odl', 'contrib')]


# Add example directories to `collect_ignore`
def find_example_dirs():
    dirs = []
    for dirpath, dirnames, _ in os.walk(odl_root):
        if 'examples' in dirnames:
            dirs.append(path.join(dirpath, 'examples'))
    return dirs


collect_ignore.extend(find_example_dirs())


if not PYFFTW_AVAILABLE:
    collect_ignore.append(
        path.join(odl_root, 'odl', 'trafos', 'backends', 'pyfftw_bindings.py')
    )
if not PYWT_AVAILABLE:
    collect_ignore.append(
        path.join(odl_root, 'odl', 'trafos', 'backends', 'pywt_bindings.py')
    )
    # Currently `pywt` is the only implementation
    collect_ignore.append(
        path.join(odl_root, 'odl', 'trafos', 'wavelet.py')
    )


# --- Command-line options --- #


def pytest_addoption(parser):
    """
    This is a pytest template to add options to the CLI when running pytest.
    It is quite handy when we want to run the test on only one backend or device, for instance.
    """
    suite_help = (
        "enable an opt-in test suite NAME. "
        "Available suites: largescale, examples, doc_doctests"
    )
    parser.addoption(
        "-S",
        "--suite",
        nargs='*',
        metavar="NAME",
        type=str,
        default=[],
        help=suite_help,
    )
    parser.addoption(
        "--backend",
        default=None,    # The default if the option is not provided
        help="Select the array backend to run the test on ('numpy'...)",
    )

    parser.addoption(
        "--device",
        default=None,    # The default if the option is not provided
        help="Select the CPU/GPU device to run the test on ('cpu'...)",
    )


def pytest_generate_tests(metafunc):
    """
    Check if the fixture name is used in a test and parametrize it
    based on the command line option.
    """
    # The name of the fixture we want to parametrize
    fixture_name = "odl_impl_device_pairs"

    if fixture_name in metafunc.fixturenames:
        # Get the CLI option
        impl   = metafunc.config.getoption("backend")
        device = metafunc.config.getoption("device")

        # Parse the backend
        if impl is not None:
            array_backend = lookup_array_backend(impl)
            if device is None:
                params=[]
                for device in array_backend.available_devices:
                    params.append((impl, device))
            else:
                device = check_device(impl, device)
                params = [(impl, device)]
        else:
            params = []
            for impl in tensor_space_impl_names():
                array_backend = lookup_array_backend(impl)
                if device is None:
                    for available_device in array_backend.available_devices:
                        params.append((impl, available_device))
                else:
                    device = check_device(impl, device)
                    params = [(impl, device)]
        
        # Parametrize the fixture. 
        # indirect=True is crucial tells pytest to pass these values 
        # to the fixture function (via request.param) rather than directly to # the test.
        metafunc.parametrize(fixture_name, params, indirect=True)


def pytest_configure(config):
    # Register an additional marker
    config.addinivalue_line(
        'markers', "suite(name): mark test to belong to an opt-in suite"
    )


def pytest_runtest_setup(item):
    suites = [mark.args[0] for mark in item.iter_markers(name='suite')]
    if suites:
        if not any(val in suites for val in item.config.getoption('-S')):
            pytest.skip(f"test not in suites {suites}")


# Remove duplicates
collect_ignore = list(set(collect_ignore))
collect_ignore = [path.normcase(ignored) for ignored in collect_ignore]


# NB: magical `path` param name is needed
def pytest_ignore_collect(path):
    """This is to ignore paths during test collection"""
    normalized = os.path.normcase(str(path))
    return any(normalized.startswith(ignored) for ignored in collect_ignore)


# --- Reusable fixtures --- #

# NOTE: All global fixtures need to be prefixed with `odl_` to make them
# non-conflicting with other packages' fixture names, since these fixtures
# are exposed globally across packages by setuptools.

# Simple ones, use helper
odl_tspace_impl = simple_fixture(name='tspace_impl',
                                 params=tensor_space_impl_names())

real_floating_dtypes = FLOAT_DTYPES
odl_real_floating_dtype = simple_fixture(name='dtype',
                                    params=real_floating_dtypes)

complex_floating_dtypes = COMPLEX_DTYPES
odl_complex_floating_dtype = simple_fixture(name='dtype',
                                    params=complex_floating_dtypes)

floating_dtypes = FLOAT_DTYPES + COMPLEX_DTYPES
odl_floating_dtype = simple_fixture(name='dtype',
                                    params=floating_dtypes)

scalar_dtypes = INTEGER_DTYPES + FLOAT_DTYPES + COMPLEX_DTYPES
odl_scalar_dtype = simple_fixture(name='dtype',
                                  params=scalar_dtypes)


@pytest.fixture(scope='module')
def odl_impl_device_pairs(request):
    """Fixture for testing accross device and backends. See
    pytest_generate_tests function to see how to modify it through the CLI
    """
    return request.param

IMPL_DEVICE_PAIRS = []
    
for impl in tensor_space_impl_names():
    array_backend = lookup_array_backend(impl)
    for device in array_backend.available_devices:
        IMPL_DEVICE_PAIRS.append((impl, device))

odl_alternative_impl_device_pair = simple_fixture(name='alternative_impl_device', params=IMPL_DEVICE_PAIRS)

if 'pytorch' in tensor_space_impl_names():
    CUDA_DEVICES = []
    for device in lookup_array_backend('pytorch').available_devices:
        CUDA_DEVICES.append(device)

    cuda_device = simple_fixture(name='cuda_device', params=CUDA_DEVICES)

odl_elem_order = simple_fixture(name='order', params=['C'])

odl_reduction = simple_fixture('reduction', ['sum', 'prod', 'min', 'max'])

# More complicated ones with non-trivial documentation
arithmetic_op_par = [operator.add,
                     operator.truediv,
                     operator.mul,
                     operator.sub,
                     operator.iadd,
                     operator.itruediv,
                     operator.imul,
                     operator.isub]
arithmetic_op_ids = [
    f" op = '{op}' " for op in ['+', '/', '*', '-', '+=', '/=', '*=', '-=']]


@fixture(ids=arithmetic_op_ids, params=arithmetic_op_par)
def odl_arithmetic_op(request):
    """An arithmetic operator, e.g. +, -, // etc."""
    return request.param



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Basic vector spaces and utilities."""

from __future__ import absolute_import

from . import base_tensors, entry_points
from .pspace import *
from .space_utils import *
from .weightings import *

__all__ = ()
__all__ += pspace.__all__
__all__ += space_utils.__all__
__all__ += weightings.__all__



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import print_function
import inspect
import importlib
import odl


__all__ = ('make_interface',)

module_string = """
.. rubric:: Modules

.. toctree::
   :maxdepth: 2

   {}
"""

fun_string = """
.. rubric:: Functions

.. autosummary::
   :toctree: generated/

   {}
"""

class_string = """
.. rubric:: Classes

.. autosummary::
   :toctree: generated/

   {}
"""

string = """{shortname}
{line}

{docstring}

.. currentmodule:: {name}

{module_string}
{class_string}
{fun_string}
"""


def import_submodules(package, name=None, recursive=True):
    """Import all submodules of ``package``.

    Parameters
    ----------
    package : `module` or string
        Package whose submodules to import.
    name : string, optional
        Override the package name with this value in the full
        submodule names. By default, ``package`` is used.
    recursive : bool, optional
        If ``True``, recursively import all submodules of ``package``.
        Otherwise, import only the modules at the top level.

    Returns
    -------
    pkg_dict : dict
        Dictionary where keys are the full submodule names and values
        are the corresponding module objects.
    """
    if isinstance(package, str):
        package = importlib.import_module(package)

    if name is None:
        name = package.__name__

    submodules = [m[0] for m in inspect.getmembers(package, inspect.ismodule)
                  if m[1].__name__.startswith('odl')]

    results = {}
    for pkgname in submodules:
        full_name = name + '.' + pkgname
        try:
            results[full_name] = importlib.import_module(full_name)
        except ImportError:
            pass
        else:
            if recursive:
                results.update(import_submodules(full_name, full_name))
    return results


def make_interface():
    """Generate the RST files for the API doc of ODL."""
    modnames = ['odl'] + list(import_submodules(odl).keys())

    for modname in modnames:
        if not modname.startswith('odl'):
            modname = 'odl.' + modname

        shortmodname = modname.split('.')[-1]
        print('{: <25} : generated {}.rst'.format(shortmodname, modname))

        line = '=' * len(shortmodname)

        module = importlib.import_module(modname)

        docstring = module.__doc__
        submodules = [m[0] for m in inspect.getmembers(
            module, inspect.ismodule) if m[1].__name__.startswith('odl')]
        functions = [m[0] for m in inspect.getmembers(
            module, inspect.isfunction) if m[1].__module__ == modname]
        classes = [m[0] for m in inspect.getmembers(
            module, inspect.isclass) if m[1].__module__ == modname]

        docstring = '' if docstring is None else docstring

        submodules = [modname + '.' + mod for mod in submodules]
        functions = ['~' + modname + '.' + fun
                     for fun in functions if not fun.startswith('_')]
        classes = ['~' + modname + '.' + cls
                   for cls in classes if not cls.startswith('_')]

        if len(submodules) > 0:
            this_mod_string = module_string.format('\n   '.join(submodules))
        else:
            this_mod_string = ''

        if len(functions) > 0:
            this_fun_string = fun_string.format('\n   '.join(functions))
        else:
            this_fun_string = ''

        if len(classes) > 0:
            this_class_string = class_string.format('\n   '.join(classes))
        else:
            this_class_string = ''

        with open(f'docs/source/{modname}.rst', 'w') as text_file:
            text_file.write(string.format(shortname=shortmodname,
                                          name=modname,
                                          line=line,
                                          docstring=docstring,
                                          module_string=this_mod_string,
                                          fun_string=this_fun_string,
                                          class_string=this_class_string))


if __name__ == '__main__':
    make_interface()



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Backend for ASTRA using CPU."""


import warnings

import numpy as np

from odl.core.discr import DiscretizedSpace, DiscretizedSpaceElement
from odl.applications.tomo.backends.astra_setup import (
    astra_algorithm, astra_data, astra_projection_geometry, astra_projector,
    astra_volume_geometry)
from odl.applications.tomo.backends.util import _add_default_complex_impl
from odl.applications.tomo.geometry import (
    DivergentBeamGeometry, Geometry, ParallelBeamGeometry)
from odl.core.util import writable_array
from odl.core.array_API_support import lookup_array_backend, get_array_and_backend

try:
    import astra
except ImportError:
    pass

__all__ = (
    'astra_cpu_projector',
    'default_astra_proj_type',
)


def default_astra_proj_type(geom):
    """Return the default ASTRA projector type for a given geometry.

    Parameters
    ----------
    geom : `Geometry`
        ODL geometry object for which to get the default projector type.

    Returns
    -------
    astra_proj_type : str
        Default projector type for the given geometry.

        In 2D:

        - `ParallelBeamGeometry`: ``'linear'``
        - `DivergentBeamGeometry`: ``'line_fanflat'``

        In 3D:

        - `ParallelBeamGeometry`: ``'linear3d'``
        - `DivergentBeamGeometry`: ``'linearcone'``
    """
    if isinstance(geom, ParallelBeamGeometry):
        return 'linear' if geom.ndim == 2 else 'linear3d'
    elif isinstance(geom, DivergentBeamGeometry):
        return 'line_fanflat' if geom.ndim == 2 else 'linearcone'
    else:
        raise TypeError(
            f"no default exists for {type(geom)}, `astra_proj_type` must be given explicitly"
        )


def astra_cpu_projector(
        direction:str,
        input_data:DiscretizedSpaceElement,
        geometry:Geometry,
        range_space:DiscretizedSpace,
        out :DiscretizedSpaceElement = None,
        astra_proj_type: str | None = None
        ) -> DiscretizedSpaceElement:
    """Run an ASTRA projection on the given data using the CPU.

    Parameters
    ----------
    input_data : `DiscretizedSpaceElement`
        Input data to which the projector is applied.
    geometry : `Geometry`
        Geometry defining the tomographic setup.
    range_space : `DiscretizedSpace`
        Space to which the calling operator maps.
    out : ``range_space`` element, optional
        Element of the range_space space to which the result is written. If
        ``None``, an element in ``range`` is created.
    astra_proj_type : str, range_space
        Type of projector that should be used. See `the ASTRA documentation
        <http://www.astra-toolbox.com/docs/proj2d.html>`_ for details.
        By default, a suitable projector type for the given geometry is
        selected, see `default_astra_proj_type`.

    Returns
    -------
    out : ``proj_space`` element
        Projection data resulting from the application of the projector.
        If ``out`` was provided, the returned object is a reference to it.
    """
    ### Asserting that we get the right data types.
    assert direction in ['forward', 'backward']
    if not isinstance(input_data, DiscretizedSpaceElement):
        raise TypeError(
            f"Input data {input_data} is not a `DiscretizedSpaceElement` instance"
        )
    if not isinstance(geometry, Geometry):
        raise TypeError(f"geometry {geometry} is not a Geometry instance")
    if not isinstance(range_space, DiscretizedSpace):
        raise TypeError(
            f"`range_space` {range_space} is not a DiscretizedSpace instance"
        )
    if input_data.ndim != geometry.ndim:
        raise ValueError(
            f"dimensions {input_data} of input data and {geometry.ndim} of geometry do not match"
        )
    if out is None:
        out_element = range_space.real_space.element()
    else:
        if out not in range_space.real_space:
            raise TypeError(
                f"`out` {out} is neither None nor a `DiscretizedSpaceElement` instance"
            )
        out_element = out.data
    ### Unpacking the dimension of the problem
    ndim = input_data.ndim
        
    ### Unpacking the underlying arrays
    input_data_arr, input_backend = get_array_and_backend(input_data, must_be_contiguous=True)

    if input_backend.impl != 'numpy':
        out_element = np.ascontiguousarray(input_backend.to_numpy(out_element))
        input_data_arr = np.ascontiguousarray(input_backend.to_numpy(input_data_arr))

    range_backend = lookup_array_backend(range_space.impl)
    assert input_backend == range_backend, f"The input's tensor space backend does not match the range's: {input_backend} != {range_backend}"    

    # Create astra geometries
    # The volume geometry is defined by the space of the input data in the forward mode and the range_space in the backward mode
    if direction == 'forward':        
        vol_geom = astra_volume_geometry(input_data.space, 'cpu')
    else:
        vol_geom = astra_volume_geometry(range_space, 'cpu')

    # Parsing the pprojection geometry does not depend on the mode       
    proj_geom = astra_projection_geometry(geometry, 'cpu')

    # Create projector
    if astra_proj_type is None:
        astra_proj_type = default_astra_proj_type(geometry)
    proj_id = astra_projector(astra_proj_type, vol_geom, proj_geom, ndim)

    # Create ASTRA data structures
    # In the forward mode, the input is the volume
    # In the backward mode, the input is the sinogram/projection
    if direction == 'forward': 
        input_id = astra_data(vol_geom, datatype='volume', data=input_data_arr,
                        allow_copy=True)
    else:
        input_id = astra_data(proj_geom, datatype='projection', data=input_data_arr, allow_copy=True
    )
    
    with writable_array(out_element, must_be_contiguous=True) as out_arr:
        if direction == 'forward': 
            output_id = astra_data(
                proj_geom, 
                datatype='projection', 
                data=out_arr,
                ndim=range_space.ndim)
            vol_id  = input_id
            sino_id = output_id
        else:
            output_id = astra_data(
                vol_geom, 
                datatype='volume', 
                data=out_arr,
                ndim=range_space.ndim)
            vol_id = output_id
            sino_id = input_id

        # Create algorithm
        algo_id = astra_algorithm(
            direction=direction,
            ndim = ndim,
            vol_id  = vol_id,
            sino_id = sino_id,
            proj_id = proj_id,
            astra_impl='cpu')

        # Run algorithm
        astra.algorithm.run(algo_id)

    # There is no scaling for the forward mode
    if direction == 'backward':
        # Weight the adjoint by appropriate weights
        scaling_factor = float(input_data.space.weighting.const)
        scaling_factor /= float(range_space.weighting.const)

        out_element *= scaling_factor

    # Delete ASTRA objects
    astra.algorithm.delete(algo_id)
    astra.data2d.delete((vol_id, sino_id))
    astra.projector.delete(proj_id)

    if out is None:
        return range_space.element(out_element)
    else:
        out.data[:] = range_space.element(out_element).data
        
class AstraCpuImpl:
    """Thin wrapper implementing ASTRA CPU for `RayTransform`."""

    def __init__(self, geometry, vol_space, proj_space):
        """Initialize a new instance.

        Parameters
        ----------
        geometry : `Geometry`
            Geometry defining the tomographic setup.
        vol_space : `DiscretizedSpace`
            Reconstruction space, the space of the images to be forward
            projected.
        proj_space : `DiscretizedSpace`
            Projection space, the space of the result.
        """
        if not isinstance(geometry, Geometry):
            raise TypeError(f"`geometry` must be a `Geometry` instance, got {geometry}")
        if not isinstance(vol_space, DiscretizedSpace):
            raise TypeError(
                f"`vol_space` must be a `DiscretizedSpace` instance, got {vol_space}"
            )
        if not isinstance(proj_space, DiscretizedSpace):
            raise TypeError(
                f"`proj_space` must be a `DiscretizedSpace` instance, got {proj_space}"
            )
        if geometry.ndim > 2:
            raise ValueError(f"`impl` {self.__class__.__name__} only works for 2d")

        if vol_space.size >= 512**2:
            warnings.warn(
                "The 'astra_cpu' backend may be too slow for volumes of this "
                "size. Consider using 'astra_cuda' if your machine has an "
                "Nvidia GPU.",
                RuntimeWarning,
            )

        self.geometry = geometry
        self._vol_space = vol_space
        self._proj_space = proj_space

    @property
    def vol_space(self):
        """Volume space of the ray transform"""
        return self._vol_space

    @property
    def proj_space(self):
        """Projection space of the ray transform"""
        return self._proj_space

    @_add_default_complex_impl
    def call_backward(self, x, out=None, **kwargs):
        """Run an ASTRA back-projection on the given data using the CPU.

        Parameters
        ----------
        proj_data : ``proj_space.real_space`` element
            Projection data to which the back-projector is applied. Although
            ``proj_space`` may be complex, this element needs to be real.
        out : ``vol_space`` element, optional
            Element of the reconstruction space to which the result is written.
            If ``None``, an element in ``vol_space`` is created.

        Returns
        -------
        out : ``vol_space`` element
            Reconstruction data resulting from the application of the
            back-projector. If ``out`` was provided, the returned object is a
            reference to it.
        """
        return astra_cpu_projector(
            'backward', x, self.geometry, self.vol_space.real_space, out, **kwargs
        )

    @_add_default_complex_impl
    def call_forward(self, x, out=None, **kwargs):
        """Run an ASTRA forward projection on the given data using the CPU.

        Parameters
        ----------
        vol_data : ``vol_space.real_space`` element
            Volume data to which the projector is applied. Although
            ``vol_space`` may be complex, this element needs to be real.
        out : ``proj_space`` element, optional
            Element of the projection space to which the result is written. If
            ``None``, an element in `proj_space` is created.

        Returns
        -------
        out : ``proj_space`` element
            Projection data resulting from the application of the projector.
            If ``out`` was provided, the returned object is a reference to it.
        """
        return astra_cpu_projector(
            'forward', x, self.geometry, self.proj_space.real_space, out, **kwargs
        )


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests

    run_doctests()



# Copyright 2014-2018 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test for parameter optimization."""

import pytest
import numpy as np
import odl
import odl.contrib.fom
import odl.contrib.param_opt
from odl.core.util.testutils import simple_fixture

space_ids =    [ "ℝ³ (float32)", "ℝ³ (float64)"
               , "uniform([0,1])", "uniform([0,1]²)" ]
space_params = [ ('ℝ³', 'float32'), ('ℝ³', 'float64')
               , ('uniform', 1), ('uniform', 2) ]

@pytest.fixture(scope="module", ids=space_ids, params=space_params)
def space(request, odl_impl_device_pairs):
    name, variation = request.param
    impl, device = 'numpy', 'cpu'
    impl, device = odl_impl_device_pairs
    if name == 'ℝ³':
        dtype = variation
        return odl.rn(3, dtype=dtype, impl=impl, device=device)
    elif name == 'uniform':
        dimension = variation
        if dimension == 1:
            return odl.uniform_discr(0, 1, 10, impl=impl, device=device)
        elif dimension == 2:
            return odl.uniform_discr([0, 0], [1, 1], [9, 11], impl=impl, device=device)
     
    raise ValueError('undefined space')


fom = simple_fixture('fom',
                     [odl.contrib.fom.mean_squared_error,
                      odl.contrib.fom.mean_absolute_error])


def test_optimal_parameters_one_parameter(space, fom):
    """Tests if optimal_parameters works for some simple examples."""
    noise = [odl.core.phantom.white_noise(space) for _ in range(2)]
    phantoms = noise.copy()
    data = noise.copy()

    def reconstruction(data, lam):
        """Simple reconstruction operator.

        The optimal reconstruction parameter in any fom is lam=0.
        """
        return 1.23 * data / lam

    result = odl.contrib.param_opt.optimal_parameters(
        reconstruction, fom, phantoms, data, [.9, 1.1], univariate=True)
    assert result == pytest.approx(1.23, abs=1e-4)


def test_optimal_parameters_two_parameters(space, fom):
    """Tests if optimal_parameters works for some simple examples."""
    noise = [odl.core.phantom.white_noise(space) for _ in range(2)]
    # Normalize to reduce test fails due to randomness
    noise = [noise_elem / noise_elem.norm() for noise_elem in noise]
    phantoms = noise.copy()
    data = noise.copy()

    def reconstruction1(data, params):
        """Perturbs the data scaling it with a nonlinear function of params."""
        return data * float((1 - params[1]) / np.abs(params[0]))

    result1 = odl.contrib.param_opt.optimal_parameters(reconstruction1, fom,
                                                       phantoms, data, [1, 2])

    tolerance = 1e-12 if space.dtype_identifier=='float64' else 1e-4

    assert 1 - result1[1] == pytest.approx(result1[0], rel=tolerance)

    def reconstruction2(data, params):
        """Perturbs the data by adding the sum of params to it."""
        return data + float(sum(params))

    result2 = odl.contrib.param_opt.optimal_parameters(reconstruction2, fom,
                                                       phantoms, data, [1, 2])
    assert sum(result2) == pytest.approx(0, abs=1e-4)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import print_function

import glob
import os
import sys

import sphinx
from packaging.version import parse as parse_version

# --- General configuration --- #

# All configuration values have a default; values that are commented out
# serve to show the default.

try:
    # Verify that we can import odl
    sys.path.insert(0, os.path.abspath('..'))
    import odl
except Exception as e:
    print('Failed importing odl, exiting', file=sys.stderr)
    print(e, file=sys.stderr)
    sys.exit(1)

import pkgutil

# -- Path setup: make sure your package is importable
sys.path.insert(0, os.path.abspath(".."))  # adjust if docs/ isn't directly inside repo root

# -- Define your package name
PACKAGE_NAME = "odl"  # <-- change this

# -- Detect available top-level modules (the ones actually installed)
installed_modules = {mod.name for mod in pkgutil.iter_modules()}

# -- Automatically mock any import that is not in installed_modules or your package
autodoc_mock_imports = ['numpy']

# # Add numpydoc path
# sys.path.insert(0, os.path.abspath('../numpydoc'))

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autosummary',
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx'
]
# Use newer 'imgmath' extension if possible
if parse_version(sphinx.__version__) >= parse_version('1.4'):
    extensions.append('sphinx.ext.imgmath')
else:
    extensions.append('sphinx.ext.pngmath')

# Add external links to GitHub
extlinks = {
    'pull': ('https://github.com/odlgroup/odl/pull/%s', 'PR %s'),
    'issue': ('https://github.com/odlgroup/odl/issues/%s', 'issue %s'),
    'commit': ('https://github.com/odlgroup/odl/commit/%s', 'commit %s')
}


# Intersphinx to get Numpy and other targets
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://docs.scipy.org/doc/numpy/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
    'matplotlib': ('https://matplotlib.org/', None),
    'pywt': ('https://pywavelets.readthedocs.io/en/latest/', None),
    'pyfftw': ('https://pyfftw.readthedocs.io/en/latest/', None),
    'pytest': ('https://docs.pytest.org/en/latest/', None)}


# Stop autodoc from skipping __init__
def skip(app, what, name, obj, skip, options):
    if (name.startswith('__') and name.endswith('__') and
        name not in ['__abstractmethods__',
                     '__doc__',
                     '__hash__',
                     '__module__',
                     '__dict__',
                     '__weakref__']):
        return False
    if name in ['_multiply',
                '_divide',
                '_lincomb',
                '_call']:
        return False
    return skip


def setup(app):
    app.connect("autodoc-skip-member", skip)
    # TODO(kohr-h): Remove when upstream issue in sphinx-rtd-theme is solved
    # https://github.com/readthedocs/sphinx_rtd_theme/issues/746
    app.add_css_file('custom.css')


# Autosummary
autosummary_generate = glob.glob("./*.rst")

# Stops WARNING: toctree contains reference to nonexisting document
# (not anymore since Sphinx 1.6)
numpydoc_show_class_members = True
numpydoc_show_inherited_class_members = True
numpydoc_class_members_toctree = True

# Set order to mirror source
autodoc_member_order = 'bysource'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'odl'
copyright = u'2014-2025 The ODL Contributors'
author = u'Jonas Adler, Holger Kohr, Justus Sagemüller, Ozan Öktem, Emilien Valat'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = odl.__version__
# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'english'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['*.py', '*.pyc']

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = 'any'

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# Warn on dead links and other stuff
nitpicky = True
nitpick_ignore = [('py:class', 'future.types.newobject.newobject')]

# --- Options for HTML output --- #

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = 'ODL'

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = 'odldoc'

# --- Options for LaTeX output --- #

latex_elements = {
    'preamble': r'''
\usepackage{amsmath}
\usepackage{amssymb}
\usepackage{enumitem}

\setlistdepth{9}
'''
}
# The paper size ('letterpaper' or 'a4paper').
# 'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
# 'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
# 'preamble': '',

# Latex figure (float) alignment
# 'figure_align': 'htbp',


# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'odl.tex', u'ODL Documentation',
     u'Jonas Adler, Holger Kohr, Justus Sagemüller, Ozan Öktem, Emilien Valat', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# --- Options for manual page output --- #

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'odl', u'ODL Documentation', [author], 1)
]

# If true, show URL addresses after external links.
# man_show_urls = False


# --- Options for Texinfo output --- #

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'odl', u'ODL Documentation',
     author, 'odl', 'One line description of project.',
     'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
# texinfo_no_detailmenu = False
